"""The `daft.catalog` module contains functionality for Catalogs.

A Catalog can be understood as a system/service for users to discover, access and query their data.
Most commonly, users' data is represented as a "table". Some more modern Catalogs such as Unity Catalog
also expose other types of data including files, ML models, registered functions and more.

Examples of Catalogs include AWS Glue, Hive Metastore, Apache Iceberg REST and Unity Catalog.

**Catalog**

```python
# without any qualifiers, the default catalog and namespace are used.
daft.read_table("my_table")

# with a qualified identifier (uses default catalog)
daft.read_table("my_namespace.my_table")

# with a fully qualified identifier
daft.read_table("my_catalog.my_namespace.my_table")
```

**Named Tables**

Daft allows also the registration of temporary tables, which have no catalog associated with them.

Note that temporary tables take precedence over catalog tables when resolving unqualified names.

```python
df = daft.from_pydict({"foo": [1, 2, 3]})

# Attach the DataFrame for use across other APIs.
daft.attach_table(df, "my_table")

# Your table is now accessible from Daft-SQL, or Daft's `read_table`
df1 = daft.read_table("my_table")
df2 = daft.sql("SELECT * FROM my_table")
```
"""

from __future__ import annotations

import warnings

from abc import ABC, abstractmethod
from collections.abc import Sequence
from daft.daft import PyIdentifier, PyTableSource

from daft.dataframe import DataFrame

from typing import TYPE_CHECKING

from daft.logical.schema import Schema

if TYPE_CHECKING:
    from daft.dataframe.dataframe import ColumnInputType


__all__ = [
    "Catalog",
    "Identifier",
    "Table",
    # TODO deprecated catalog APIs #3819
    "read_table",
    "register_python_catalog",
    "register_table",
    "unregister_catalog",
]


# TODO deprecated catalog APIs #3819
def unregister_catalog(catalog_name: str | None) -> bool:
    """Unregisters a catalog from the Daft catalog system.

    DEPRECATED: This is deprecated and will be removed in daft >= 0.5.0; please use `daft.detach_catalog`.

    This function removes a previously registered catalog from the Daft catalog system.

    Args:
        catalog_name (Optional[str]): The name of the catalog to unregister. If None, the default catalog will be unregistered.

    Returns:
        bool: True if a catalog was successfully unregistered, False otherwise.

    Example:
        >>> import daft
        >>> daft.unregister_catalog("my_catalog")
        True
    """
    from daft.session import detach_catalog

    warnings.warn(
        "This is deprecated and will be removed in daft >= 0.5.0; please use `daft.detach_catalog`.",
        category=DeprecationWarning,
    )
    try:
        alias = catalog_name if catalog_name else "default"
        detach_catalog(alias)
        return True
    except Exception:
        return False


# TODO deprecated catalog APIs #3819
def read_table(name: str) -> DataFrame:
    """Finds a table with the specified name and reads it as a DataFrame.

    DEPRECATED: This is deprecated and will be removed in daft >= 0.5.0; please use `daft.read_table`.

    The provided name can be any of the following, and Daft will return them with the following order of priority:

    1. Name of a registered dataframe/SQL view (manually registered using `daft.register_table`): `"my_registered_table"`
    2. Name of a table within the default catalog (without inputting the catalog name) for example: `"my.table.name"`
    3. Name of a fully-qualified table path with the catalog name for example: `"my_catalog.my.table.name"`

    Args:
        name: The identifier for the table to read

    Returns:
        A DataFrame containing the data from the specified table.
    """
    from daft.session import read_table

    warnings.warn(
        "This is deprecated and will be removed in daft >= 0.5.0; please use `daft.read_table`.",
        category=DeprecationWarning,
    )
    return read_table(name)


# TODO deprecated catalog APIs #3819
def register_table(name: str, dataframe: DataFrame) -> str:
    """Register a DataFrame as a named table.

    DEPRECATED: This is deprecated and will be removed in daft >= 0.5.0; please use `daft.attach_table`.

    This function registers a DataFrame as a named table, making it accessible
    via Daft-SQL or Daft's `read_table` function.

    Args:
        name (str): The name to register the table under.
        dataframe (daft.DataFrame): The DataFrame to register as a table.

    Returns:
        str: The name of the registered table.

    Example:
        >>> df = daft.from_pydict({"foo": [1, 2, 3]})
        >>> daft.catalog.register_table("my_table", df)
        >>> daft.read_table("my_table")
    """
    from daft.session import create_temp_table

    warnings.warn(
        "This is deprecated and will be removed in daft >= 0.5.0; please use `daft.create_temp_table`.",
        category=DeprecationWarning,
    )
    _ = create_temp_table(name, dataframe)
    return name


# TODO deprecated catalog APIs #3819
def register_python_catalog(catalog: object, name: str | None = None) -> str:
    """Registers a Python catalog with Daft.

    DEPRECATED: This is deprecated and will be removed in daft >= 0.5.0; please use `daft.attach_catalog`.

    Currently supports:

    * [PyIceberg Catalogs](https://py.iceberg.apache.org/api/)
    * [Unity Catalog](https://www.getdaft.io/projects/docs/en/latest/user_guide/integrations/unity-catalog.html)

    Args:
        catalog (PyIcebergCatalog | UnityCatalog): The Python catalog to register.
        name (str | None, optional): The name to register the catalog under. If None, this catalog is registered as the default catalog.

    Returns:
        str: The name of the registered catalog.

    Raises:
        ValueError: If an unsupported catalog type is provided.

    Example:
        >>> from pyiceberg.catalog import load_catalog
        >>> catalog = load_catalog("my_catalog")
        >>> daft.catalog.register_python_catalog(catalog, "my_daft_catalog")

    """
    from daft.session import attach_catalog

    warnings.warn(
        "This is deprecated and will be removed in daft >= 0.5.0; please use `daft.attach_catalog`.",
        category=DeprecationWarning,
    )
    if name is None:
        name = "default"
    _ = attach_catalog(catalog, name)
    return name


class Catalog(ABC):
    """Interface for python catalog implementations."""

    @staticmethod
    def from_pydict(tables: dict[str, Table]) -> Catalog:
        """Returns an in-memory catalog from the dictionary."""
        from daft.catalog.__memory import MemoryCatalog

        return MemoryCatalog(tables)

    @staticmethod
    def from_iceberg(obj: object) -> Catalog:
        """Returns a Daft Catalog instance from an Iceberg catalog."""
        try:
            from daft.catalog.__iceberg import IcebergCatalog

            return IcebergCatalog._from_obj(obj)
        except ImportError:
            raise ImportError("Iceberg support not installed: pip install -U 'getdaft[iceberg]'")

    @staticmethod
    def from_unity(obj: object) -> Catalog:
        """Returns a Daft Catalog instance from a Unity catalog."""
        try:
            from daft.catalog.__unity import UnityCatalog

            return UnityCatalog._from_obj(obj)
        except ImportError:
            raise ImportError("Unity support not installed: pip install -U 'getdaft[unity]'")

    @staticmethod
    def _from_obj(obj: object) -> Catalog:
        """Returns a Daft Catalog from a supported object type or raises a ValueError."""
        for factory in (Catalog.from_iceberg, Catalog.from_unity):
            try:
                return factory(obj)
            except ValueError:
                pass
            except ImportError:
                pass
        raise ValueError(
            f"Unsupported catalog type: {type(obj)}; please ensure all required extra dependencies are installed."
        )

    ###
    # list_*
    ###

    @abstractmethod
    def list_tables(self, pattern: str | None = None) -> list[str]: ...

    ###
    # get_*
    ###

    @abstractmethod
    def get_table(self, identifier: Identifier | str) -> Table: ...

    # TODO deprecated catalog APIs #3819
    def load_table(self, name: str) -> Table:
        """DEPRECATED: Please use `get_table` instead; version=0.5.0!"""
        warnings.warn(
            "This is deprecated and will be removed in daft >= 0.5.0, please use `get_table` instead.",
            category=DeprecationWarning,
        )
        return self.get_table(name)


class Identifier(Sequence):
    """A reference (path) to a catalog object.

    Example:
    >>> id = Identifier("a", "b")
    >>> assert len(id) == 2
    """

    _ident: PyIdentifier

    def __init__(self, *parts: str):
        """Creates an Identifier from its parts.

        Example:
        >>> Identifier("namespace", "table")

        Returns:
            Identifier: A new identifier.
        """
        if len(parts) < 1:
            raise ValueError("Identifier requires at least one part.")
        self._ident = PyIdentifier(parts[:-1], parts[-1])

    @staticmethod
    def _from_pyidentifier(ident: PyIdentifier) -> Identifier:
        i = Identifier.__new__(Identifier)
        i._ident = ident
        return i

    @staticmethod
    def from_sql(input: str, normalize: bool = False) -> Identifier:
        """Parses an Identifier from an SQL string, normalizing to lowercase if specified.

        Example:
        >>> Identifier.from_sql("namespace.table") == Identifier("namespace", "table")
        >>> Identifier.from_sql('"a.b"') == Identifier('"a.b."')
        >>> Identifier.from_sql('ABC."xYz"', normalize=True) == Identifier("abc", "xYz")

        Returns:
            Identifier: A new identifier.
        """
        i = Identifier.__new__(Identifier)
        i._ident = PyIdentifier.from_sql(input, normalize)
        return i

    @staticmethod
    def from_str(input: str) -> Identifier:
        """Parses an Identifier from a dot-delimited Python string without normalization."""
        return Identifier(*input.split("."))

    def __eq__(self, other: object) -> bool:
        if not isinstance(other, Identifier):
            return False
        return self._ident.eq(other._ident)

    def __getitem__(self, index: int | slice) -> str | Sequence[str]:
        if isinstance(index, slice):
            raise IndexError("slicing not supported")
        if isinstance(index, int):
            return self._ident.getitem(index)

    def __len__(self) -> int:
        return self._ident.__len__()

    def __repr__(self) -> str:
        return f"Identifier('{self._ident.__repr__()}')"

    def __str__(self) -> str:
        return ".".join(self)


class Table(ABC):
    """Interface for python table implementations."""

    @staticmethod
    def from_df(name: str, dataframe: DataFrame) -> Table:
        """Returns a read-only table backed by the DataFrame."""
        from daft.catalog.__memory import MemoryTable

        return MemoryTable(name, dataframe)

    @staticmethod
    def from_iceberg(obj: object) -> Table:
        """Returns a Daft Table instance from an Iceberg table."""
        try:
            from daft.catalog.__iceberg import IcebergTable

            return IcebergTable._from_obj(obj)
        except ImportError:
            raise ImportError("Iceberg support not installed: pip install -U 'getdaft[iceberg]'")

    @staticmethod
    def from_unity(obj: object) -> Table:
        """Returns a Daft Table instance from a Unity table."""
        try:
            from daft.catalog.__unity import UnityTable

            return UnityTable._from_obj(obj)
        except ImportError:
            raise ImportError("Unity support not installed: pip install -U 'getdaft[unity]'")

    @staticmethod
    def _from_obj(obj: object) -> Table:
        """Returns a Daft Table from a supported object type or raises an error."""
        raise ValueError(f"Unsupported table type: {type(obj)}")

    # TODO catalog APIs part 3
    # @property
    # @abstractmethod
    # def name(self) -> str:
    #     """Returns the table name."""

    # TODO catalog APIs part 3
    # @property
    # @abstractmethod
    # def inner(self) -> object | None:
    #     """Returns the inner table object if this is an adapter."""

    @abstractmethod
    def read(self) -> DataFrame:
        """Returns a DataFrame from this table."""

    # TODO deprecated catalog APIs #3819
    def to_dataframe(self) -> DataFrame:
        """DEPRECATED: Please use `read` instead; version 0.5.0!"""
        warnings.warn(
            "This is deprecated and will be removed in daft >= 0.5.0, please use `read` instead.",
            category=DeprecationWarning,
        )
        return self.read()

    def select(self, *columns: ColumnInputType) -> DataFrame:
        """Returns a DataFrame from this table with the selected columns."""
        return self.read().select(*columns)

    def show(self, n: int = 8) -> None:
        """Shows the first n rows from this table."""
        return self.read().show(n)


class TableSource:
    """A TableSource is used to create a new table; this could be a Schema or DataFrame."""

    _source: PyTableSource

    def __init__(self) -> None:
        raise ValueError("We do not support creating a TableSource via __init__")

    @staticmethod
    def from_df(df: DataFrame) -> TableSource:
        s = TableSource.__new__(TableSource)
        s._source = PyTableSource.from_builder(df._builder._builder)
        return s

    @staticmethod
    def _from_obj(obj: object = None) -> TableSource:
        # TODO for future sources, consider https://github.com/Eventual-Inc/Daft/pull/2864
        if obj is None:
            return TableSource._from_none()
        elif isinstance(obj, DataFrame):
            return TableSource.from_df(obj)
        elif isinstance(obj, str):
            return TableSource._from_path(obj)
        elif isinstance(obj, Schema):
            return TableSource._from_schema(obj)
        else:
            raise Exception(f"Unknown table source: {obj}")

    @staticmethod
    def _from_none() -> TableSource:
        # for creating temp mutable tables, but we don't have those yet
        # s = TableSource.__new__(TableSource)
        # s._source = PyTableSource.empty()
        # return s
        # todo temp workaround just use an empty schema
        return TableSource._from_schema(Schema._from_fields([]))

    @staticmethod
    def _from_schema(schema: Schema) -> TableSource:
        # we don't have mutable temp tables, so just make an empty view
        # s = TableSource.__new__(TableSource)
        # s._source = PyTableSource.from_schema(schema._schema)
        # return s
        # todo temp workaround until create_table is wired
        return TableSource.from_df(DataFrame._from_pylist([]))

    @staticmethod
    def _from_path(path: str) -> TableSource:
        # for supporting daft.create_table("t", "/path/to/data") <-> CREATE TABLE t AS '/path/to/my.data'
        raise NotImplementedError("creating a table source from a path is not yet supported.")



import pytest

from daft import Identifier


def assert_eq(parsed_id, expect_id):
    assert parsed_id == expect_id


def test_identifier_regular_single():
    input_sql = "a"
    parsed_id = Identifier.from_sql(input_sql)
    expect_id = Identifier("a")
    assert_eq(parsed_id, expect_id)


def test_identifier_regular_multi():
    input_sql = "a.b"
    parsed_id = Identifier.from_sql(input_sql)
    expect_id = Identifier("a", "b")
    assert_eq(parsed_id, expect_id)


def test_identifier_delimited_single():
    input_sql = """ "a" """
    parsed_id = Identifier.from_sql(input_sql)
    expect_id = Identifier("a")
    assert_eq(parsed_id, expect_id)


def test_identifier_delimited_multi():
    input_sql = """ "a"."b" """
    parsed_id = Identifier.from_sql(input_sql)
    expect_id = Identifier("a", "b")
    assert_eq(parsed_id, expect_id)


def test_identifier_mixed():
    input_sql = """ "a".b."c" """
    parsed_id = Identifier.from_sql(input_sql)
    expect_id = Identifier("a", "b", "c")
    assert_eq(parsed_id, expect_id)


def test_identifier_special():
    input_sql = """ "a.b"."-^-" """
    parsed_id = Identifier.from_sql(input_sql)
    expect_id = Identifier("a.b", "-^-")
    assert_eq(parsed_id, expect_id)


def test_identifier_sequence():
    parts = ["a", "b", "c"]
    ident = Identifier(*parts)
    # __len__
    assert len(parts) == len(ident)
    # __getitem__
    for i in range(len(parts)):
        assert parts[i] == ident[i]
        assert parts[-i] == ident[-i]
    # __iter__ (derived)
    for i, part in enumerate(ident):
        assert parts[i] == part


@pytest.mark.parametrize(
    "input",
    [
        "1",  # can't start with numbers
        "a.",  # invalid: .
        "a b",  # invalid: <space>
        "a-b",  # invalid: -
        "'a'",  # string literal, not an ident
        "`a`",  # backtick non-ANSI.
    ],
)
def test_invalid_identifier(input):
    with pytest.raises(Exception, match="Invalid identifier"):
        Identifier.from_sql(input)


# ensure we can support the customer ask: https://github.com/Eventual-Inc/Daft/issues/3758
def test_identifier_with_periods():
    assert 3 == len(Identifier.from_sql("a.b.c"))
    assert 1 == len(Identifier.from_sql('"a.b.c"'))



from __future__ import annotations

import itertools
import math
import operator as ops

import numpy as np
import pyarrow as pa
import pytest

from daft import col, lit
from daft.recordbatch import MicroPartition
from tests.recordbatch import daft_numeric_types

OPS = [
    ops.add,
    ops.sub,
    ops.mul,
    ops.truediv,
    ops.floordiv,
    ops.mod,
    ops.lt,
    ops.le,
    ops.eq,
    ops.ne,
    ops.ge,
    ops.gt,
]


def lists_close_with_nones(a, b):
    if len(a) != len(b):
        return False
    for x, y in zip(a, b):
        if x is None and y is None:
            continue
        if x is not None and y is not None:
            np.testing.assert_allclose([x], [y])
        else:
            return False
    return True


@pytest.mark.parametrize("data_dtype, op", itertools.product(daft_numeric_types, OPS))
def test_table_numeric_expressions(data_dtype, op) -> None:
    a, b = [5, 6, 7, 8], [1, 2, 3, 4]
    pa_table = pa.Table.from_pydict({"a": a, "b": b})

    daft_recordbatch = MicroPartition.from_arrow(pa_table)
    daft_recordbatch = daft_recordbatch.eval_expression_list(
        [op(col("a").cast(data_dtype), col("b").cast(data_dtype)).alias("result")]
    )

    assert len(daft_recordbatch) == 4
    assert daft_recordbatch.column_names() == ["result"]
    pyresult = [op(left, right) for left, right in zip(a, b)]
    assert daft_recordbatch.get_column("result").to_pylist() == pyresult


@pytest.mark.parametrize("data_dtype, op", itertools.product(daft_numeric_types, OPS))
def test_table_numeric_expressions_with_nulls(data_dtype, op) -> None:
    a, b = [5, 6, None, 8, None], [1, 2, 3, None, None]
    pa_table = pa.Table.from_pydict({"a": a, "b": b})

    daft_recordbatch = MicroPartition.from_arrow(pa_table)
    daft_recordbatch = daft_recordbatch.eval_expression_list(
        [op(col("a").cast(data_dtype), col("b").cast(data_dtype)).alias("result")]
    )

    assert len(daft_recordbatch) == 5
    assert daft_recordbatch.column_names() == ["result"]
    pyresult = [op(left, right) for left, right in zip(a[:2], b[:2])]
    assert daft_recordbatch.get_column("result").to_pylist()[:2] == pyresult

    assert daft_recordbatch.get_column("result").to_pylist()[2:] == [None, None, None]


def test_table_numeric_abs() -> None:
    table = MicroPartition.from_pydict({"a": [None, -1.0, 0, 2, 3, None], "b": [-1, -2, 3, 4, None, None]})

    abs_table = table.eval_expression_list([abs(col("a")), col("b").abs()])

    assert [abs(v) if v is not None else v for v in table.get_column("a").to_pylist()] == abs_table.get_column(
        "a"
    ).to_pylist()
    assert [abs(v) if v is not None else v for v in table.get_column("b").to_pylist()] == abs_table.get_column(
        "b"
    ).to_pylist()


def test_table_abs_bad_input() -> None:
    table = MicroPartition.from_pydict({"a": ["a", "b", "c"]})

    with pytest.raises(ValueError, match="Expected input to abs to be numeric"):
        table.eval_expression_list([abs(col("a"))])


def test_table_numeric_ceil() -> None:
    table = MicroPartition.from_pydict(
        {
            "a": [None, -1.0, -0.5, 0, 0.5, 2, None],
            "b": [-1.7, -1.5, -1.3, 0.3, 0.7, None, None],
        }
    )

    ceil_table = table.eval_expression_list([col("a").ceil(), col("b").ceil()])

    assert [math.ceil(v) if v is not None else v for v in table.get_column("a").to_pylist()] == ceil_table.get_column(
        "a"
    ).to_pylist()
    assert [math.ceil(v) if v is not None else v for v in table.get_column("b").to_pylist()] == ceil_table.get_column(
        "b"
    ).to_pylist()


def test_table_ceil_bad_input() -> None:
    table = MicroPartition.from_pydict({"a": ["a", "b", "c"]})

    with pytest.raises(ValueError, match="Expected input to ceil to be numeric"):
        table.eval_expression_list([col("a").ceil()])


def test_table_numeric_floor() -> None:
    table = MicroPartition.from_pydict(
        {
            "a": [None, -1.0, -0.5, 0.0, 0.5, 2, None],
            "b": [-1.7, -1.5, -1.3, 0.3, 0.7, None, None],
        }
    )

    floor_table = table.eval_expression_list([col("a").floor(), col("b").floor()])

    assert [math.floor(v) if v is not None else v for v in table.get_column("a").to_pylist()] == floor_table.get_column(
        "a"
    ).to_pylist()
    assert [math.floor(v) if v is not None else v for v in table.get_column("b").to_pylist()] == floor_table.get_column(
        "b"
    ).to_pylist()


def test_table_floor_bad_input() -> None:
    table = MicroPartition.from_pydict({"a": ["a", "b", "c"]})

    with pytest.raises(ValueError, match="Expected input to floor to be numeric"):
        table.eval_expression_list([col("a").floor()])


def test_table_numeric_sign() -> None:
    table = MicroPartition.from_pydict(
        {
            "a": [None, -1, -5, 0, 5, 2, None],
            "b": [-1.7, -1.5, -1.3, 0.3, 0.7, None, None],
        }
    )
    my_schema = pa.schema([pa.field("uint8", pa.uint8())])
    table_Unsign = MicroPartition.from_arrow(pa.Table.from_arrays([pa.array([None, 0, 1, 2, 3])], schema=my_schema))

    sign_table = table.eval_expression_list([col("a").sign(), col("b").sign()])
    unsign_sign_table = table_Unsign.eval_expression_list([col("uint8").sign()])

    def checkSign(val):
        if val < 0:
            return -1
        if val > 0:
            return 1
        return 0

    assert [checkSign(v) if v is not None else v for v in table.get_column("a").to_pylist()] == sign_table.get_column(
        "a"
    ).to_pylist()
    assert [checkSign(v) if v is not None else v for v in table.get_column("b").to_pylist()] == sign_table.get_column(
        "b"
    ).to_pylist()
    assert [
        checkSign(v) if v is not None else v for v in table_Unsign.get_column("uint8").to_pylist()
    ] == unsign_sign_table.get_column("uint8").to_pylist()


def test_table_sign_bad_input() -> None:
    table = MicroPartition.from_pydict({"a": ["a", "b", "c"]})

    with pytest.raises(ValueError, match="Expected input to sign to be numeric"):
        table.eval_expression_list([col("a").sign()])


@pytest.mark.parametrize(
    ("fun", "is_arc"),
    [
        ("sin", False),
        ("cos", False),
        ("tan", False),
        ("arcsin", True),
        ("arccos", True),
        ("arctan", True),
        ("radians", False),
        ("degrees", False),
    ],
)
def test_table_numeric_trigonometry(fun: str, is_arc: bool) -> None:
    if not is_arc:
        table = MicroPartition.from_pydict({"a": [0.0, math.pi, math.pi / 2, math.nan]})
    else:
        table = MicroPartition.from_pydict({"a": [0.0, 1, 0.5, math.nan]})
    s = table.to_pandas()["a"]
    np_result = getattr(np, fun)(s)

    trigonometry_table = table.eval_expression_list([getattr(col("a"), fun)()])
    assert (
        all(
            x == y or (math.isnan(x) and math.isnan(y))
            for x, y in zip(trigonometry_table.get_column("a").to_pylist(), np_result.to_list())
        )
        is True
    )


def test_table_numeric_arc_trigonometry_oor() -> None:
    table = MicroPartition.from_pydict({"a": [math.pi, 2]})
    cot_table = table.eval_expression_list([col("a").arcsin(), col("a").arccos().alias("b")])
    assert all(math.isnan(x) for x in cot_table.get_column("a").to_pylist())
    assert all(math.isnan(x) for x in cot_table.get_column("b").to_pylist())


def test_table_numeric_cot() -> None:
    table = MicroPartition.from_pydict({"a": [0, None, math.nan]})
    cot_table = table.eval_expression_list([col("a").cot()])
    expected = [math.inf, None, math.nan]
    assert (
        all(
            x == y or (math.isnan(x) and math.isnan(y)) or (math.isinf(x) and math.isinf(y))
            for x, y in zip(cot_table.get_column("a").to_pylist(), expected)
        )
        is True
    )


def test_table_numeric_atan2() -> None:
    # cartesian product of y and x tables
    table = MicroPartition.from_pydict(
        {
            "y": [0.0, 1.0, 0.5, -0.5, -0.0, math.nan, 0.0, math.nan],
            "x": [0.0, 0.0, 0.5, 0.5, -10.0, math.nan, math.nan, 1.0],
        }
    )
    pds = table.to_pandas()
    np_result = np.arctan2(pds["y"], pds["x"])

    atan2_table = table.eval_expression_list([col("y").arctan2(col("x"))])
    assert (
        all(
            a == b or (a is None and b is None) or (math.isnan(a) and math.isnan(b))
            for a, b in zip(atan2_table.get_column("y").to_pylist(), np_result.to_list())
        )
        is True
    )


def test_table_numeric_atan2_literals() -> None:
    table = MicroPartition.from_pydict({"y": [0.0, 1.0, -1.0, math.nan]})
    pds = table.to_pandas()
    literals = [0.0, 1.0, -1.0, math.nan]
    # lhs has value, rhs has literal
    for litv in literals:
        np_result = np.arctan2(pds["y"], np.repeat(litv, len(pds)))
        atan2_table = table.eval_expression_list([col("y").arctan2(lit(litv))])
        assert (
            all(
                a == b or (a is None and b is None) or (math.isnan(a) and math.isnan(b))
                for a, b in zip(atan2_table.get_column("y").to_pylist(), np_result.to_list())
            )
            is True
        )

    # lhs has literal, rhs has value
    for litv in literals:
        np_result = np.arctan2(np.repeat(litv, len(pds)), pds["y"])
        atan2_table = table.eval_expression_list([lit(litv).arctan2(col("y"))])
        assert (
            all(
                a == b or (a is None and b is None) or (math.isnan(a) and math.isnan(b))
                for a, b in zip(atan2_table.get_column("literal").to_pylist(), np_result.to_list())
            )
            is True
        )


def test_table_numeric_arctanh() -> None:
    table = MicroPartition.from_pydict({"a": [0.0, 0.5, 0.9, -0.9, -0.5, -0.0, 1, -1.3, math.nan]})
    s = table.to_pandas()["a"]
    np_result = np.arctanh(s)

    arct = table.eval_expression_list([col("a").arctanh()])
    assert (
        all(
            x - y < 1.0e-10
            or (x is None and y is None)
            or (math.isnan(x) and math.isnan(y) or math.isinf(x) and math.isinf(y))
            for x, y in zip(arct.get_column("a").to_pylist(), np_result.to_list())
        )
        is True
    )


def test_table_numeric_arcsinh() -> None:
    table = MicroPartition.from_pydict({"a": [0.0, 1.0, 0.5, -0.5, -0.0, math.nan]})
    s = table.to_pandas()["a"]
    np_result = np.arcsinh(s)

    arcs = table.eval_expression_list([col("a").arcsinh()])
    assert (
        all(
            x - y < 1.0e-10 or (x is None and y is None) or (math.isnan(x) and math.isnan(y))
            for x, y in zip(arcs.get_column("a").to_pylist(), np_result.to_list())
        )
        is True
    )


def test_table_numeric_arccosh() -> None:
    table = MicroPartition.from_pydict({"a": [1.0, 2.0, 1.5, 0.5, math.nan]})
    s = table.to_pandas()["a"]
    np_result = np.arccosh(s)

    arcc = table.eval_expression_list([col("a").arccosh()])
    assert (
        all(
            x - y < 1.0e-10 or (x is None and y is None) or (math.isnan(x) and math.isnan(y))
            for x, y in zip(arcc.get_column("a").to_pylist(), np_result.to_list())
        )
        is True
    )


def test_table_numeric_round() -> None:
    from decimal import ROUND_HALF_UP, Decimal

    table = MicroPartition.from_pydict(
        {
            "a": [None, -1, -5, 0, 5, 2, None],
            "b": [-1.765, -1.565, -1.321, 0.399, 0.781, None, None],
        }
    )
    round_table = table.eval_expression_list([col("a").round(0), col("b").round(2)])
    assert [
        Decimal(v).to_integral_value(rounding=ROUND_HALF_UP) if v is not None else v
        for v in table.get_column("a").to_pylist()
    ] == round_table.get_column("a").to_pylist()
    assert [
        (float(Decimal(str(v)).quantize(Decimal("0.00"), rounding=ROUND_HALF_UP)) if v is not None else v)
        for v in table.get_column("b").to_pylist()
    ] == round_table.get_column("b").to_pylist()


def test_table_round_bad_input() -> None:
    table = MicroPartition.from_pydict({"a": ["a", "b", "c"]})

    with pytest.raises(ValueError, match="Expected input to round to be numeric"):
        table.eval_expression_list([col("a").round()])

    table = MicroPartition.from_pydict({"a": [1, 2, 3]})

    with pytest.raises(ValueError, match="decimal can not be negative: -2"):
        table.eval_expression_list([col("a").round(-2)])


def test_clip_one_sided_bounding():
    table = MicroPartition.from_pydict({"a": [1, 2, 3, 4, 5]})
    clip_table = table.eval_expression_list([col("a").clip(None, 3)])
    expected = [1, 2, 3, 3, 3]
    assert clip_table.get_column("a").to_pylist() == expected

    clip_table = table.eval_expression_list([col("a").clip(3, None)])
    expected = [3, 3, 3, 4, 5]
    assert clip_table.get_column("a").to_pylist() == expected


def test_clip_integer_float_promotion():
    table = MicroPartition.from_pydict({"a": [1, 2, 3, 4, 5]})
    clip_table = table.eval_expression_list([col("a").clip(2.5, 4.5)])
    expected = [2.5, 2.5, 3, 4, 4.5]
    assert clip_table.get_column("a").to_pylist() == expected


def test_clip_zero_handling():
    table = MicroPartition.from_pydict({"a": [-0.0, 0.0, 1.0, -1.0]})
    clip_table = table.eval_expression_list([col("a").clip(-0.5, 0.5)])
    expected = [-0.0, 0.0, 0.5, -0.5]
    assert clip_table.get_column("a").to_pylist() == expected


def test_clip_empty_array():
    table = MicroPartition.from_pydict({"a": []})
    with pytest.raises(ValueError):
        table.eval_expression_list([col("a").clip(0, 1)])


def test_clip_all_within_bounds():
    table = MicroPartition.from_pydict({"a": [2, 3, 4]})
    clip_table = table.eval_expression_list([col("a").clip(1, 5)])
    expected = [2, 3, 4]
    assert clip_table.get_column("a").to_pylist() == expected


def test_clip_all_out_of_bounds():
    table = MicroPartition.from_pydict({"a": [0, 1, 2, 3, 4, 5]})
    clip_table = table.eval_expression_list([col("a").clip(2, 3)])
    expected = [2, 2, 2, 3, 3, 3]
    assert clip_table.get_column("a").to_pylist() == expected


def test_clip_nan_handling():
    table = MicroPartition.from_pydict({"a": [1, 2, np.nan, 4, 5]})
    clip_table = table.eval_expression_list([col("a").clip(2, 4)])
    expected = [2, 2, np.nan, 4, 4]
    actual = clip_table.get_column("a").to_pylist()
    assert all((a == b or (np.isnan(a) and np.isnan(b))) for a, b in zip(actual, expected))


@pytest.mark.parametrize(
    "lower_bound, upper_bound, expected",
    [
        # Test case 1: Column lower bound with scalar upper bound
        ("lower_bound", 5, [1.0, 2.5, None, 4.7, 5.0, float("nan")]),
        # Test case 2: Scalar lower bound with column upper bound
        (2.0, "upper_bound", [2.0, 2.5, None, 4.7, 5.0, float("nan")]),
        # Test case 3: Column lower and upper bounds
        ("lower_bound", "upper_bound", [1.0, 2.5, None, 4.7, 5.0, float("nan")]),
        # Test case 4: Infinite bounds
        (float("-inf"), float("inf"), [1.0, 2.5, None, 4.7, 5.0, float("nan")]),
        # Test case 5: None bounds
        (None, None, [1.0, 2.5, None, 4.7, 5.0, float("nan")]),
        # Test case 6: Scalar bounds
        (2.0, 5.0, [2.0, 2.5, None, 4.7, 5.0, float("nan")]),
    ],
)
def test_clip(lower_bound, upper_bound, expected):
    """Test clipping a column with various combinations of scalar and column bounds."""
    # Initialize test data
    table = MicroPartition.from_pydict(
        {
            "data": [1.0, 2.5, None, 4.7, 5.0, float("nan")],
            "lower_bound": [0.5, 2.0, 1.0, None, 4.0, 0.0],
            "upper_bound": [2.0, 3.0, 5.0, None, None, float("inf")],
        }
    )

    # Prepare the clip expression
    lower = col(lower_bound) if isinstance(lower_bound, str) else lower_bound
    upper = col(upper_bound) if isinstance(upper_bound, str) else upper_bound

    # Perform the clip operation
    clip_table = table.eval_expression_list([col("data").clip(lower, upper)])
    actual = clip_table.get_column("data").to_pylist()

    # Verify results
    assert all(
        (a == b) or (a is None and b is None) or (math.isnan(a) and math.isnan(b)) for a, b in zip(actual, expected)
    ), f"Expected {expected}, got {actual}"


def test_clip_incompatible_lengths():
    table1 = MicroPartition.from_pydict({"data": [1, 2, 3, 4, 5]})
    table2 = MicroPartition.from_pydict({"data": [1, 2, 3]})  # Shorter array
    table3 = MicroPartition.from_pydict({"data": [1, 2, 3, 4, 5, 6]})  # Longer array

    # Test shorter lower bound
    with pytest.raises(ValueError):
        table1.eval_expression_list([col("data").clip(table2.get_column("data"), lit(5))])

    # Test longer upper bound
    with pytest.raises(ValueError):
        table1.eval_expression_list([col("data").clip(lit(0), table3.get_column("data"))])

    # Test both bounds with different lengths
    with pytest.raises(ValueError):
        table1.eval_expression_list([col("data").clip(table2.get_column("data"), table3.get_column("data"))])


def test_clip_unsupported_types():
    # Test with string data
    table = MicroPartition.from_pydict({"strings": ["a", "b", "c"], "numbers": [1, 2, 3]})

    # Test string column as data
    with pytest.raises(ValueError):
        table.eval_expression_list([col("strings").clip(0, 5)])

    # Test string column as lower bound
    with pytest.raises(ValueError):
        table.eval_expression_list([col("numbers").clip(col("strings"), 5)])

    # Test string column as upper bound
    with pytest.raises(ValueError):
        table.eval_expression_list([col("numbers").clip(0, col("strings"))])

    # Test string literals as bounds
    with pytest.raises(ValueError):
        table.eval_expression_list([col("numbers").clip("a", "z")])


def test_table_numeric_log2() -> None:
    table = MicroPartition.from_pydict({"a": [0.1, 0.01, 1.5, None], "b": [1, 10, None, None]})
    log2_table = table.eval_expression_list([col("a").log2(), col("b").log2()])
    assert [math.log2(v) if v is not None else v for v in table.get_column("a").to_pylist()] == log2_table.get_column(
        "a"
    ).to_pylist()
    assert [math.log2(v) if v is not None else v for v in table.get_column("b").to_pylist()] == log2_table.get_column(
        "b"
    ).to_pylist()


def test_table_log2_bad_input() -> None:
    table = MicroPartition.from_pydict({"a": ["a", "b", "c"]})

    with pytest.raises(ValueError, match="Expected input to log to be numeric"):
        table.eval_expression_list([col("a").log2()])


def test_table_numeric_log10() -> None:
    table = MicroPartition.from_pydict({"a": [0.1, 0.01, 1.5, None], "b": [1, 10, None, None]})
    log10_table = table.eval_expression_list([col("a").log10(), col("b").log10()])
    assert [math.log10(v) if v is not None else v for v in table.get_column("a").to_pylist()] == log10_table.get_column(
        "a"
    ).to_pylist()
    assert [math.log10(v) if v is not None else v for v in table.get_column("b").to_pylist()] == log10_table.get_column(
        "b"
    ).to_pylist()


def test_table_log10_bad_input() -> None:
    table = MicroPartition.from_pydict({"a": ["a", "b", "c"]})

    with pytest.raises(ValueError, match="Expected input to log to be numeric"):
        table.eval_expression_list([col("a").log10()])


@pytest.mark.parametrize(
    ("base"),
    [2, 10, 100, math.e],
)
def test_table_numeric_log(base: float) -> None:
    table = MicroPartition.from_pydict({"a": [0.1, 0.01, 1.5, None], "b": [1, 10, None, None]})
    log_table = table.eval_expression_list([col("a").log(base), col("b").log(base)])

    assert lists_close_with_nones(
        log_table.get_column("a").to_pylist(),
        [math.log(v, base) if v is not None else None for v in table.get_column("a").to_pylist()],
    )

    assert lists_close_with_nones(
        log_table.get_column("b").to_pylist(),
        [math.log(v, base) if v is not None else None for v in table.get_column("b").to_pylist()],
    )


def test_table_log_bad_input() -> None:
    table = MicroPartition.from_pydict({"a": ["a", "b", "c"]})

    with pytest.raises(ValueError, match="Expected input to log to be numeric"):
        table.eval_expression_list([col("a").log(base=10)])


def test_table_numeric_ln() -> None:
    table = MicroPartition.from_pydict({"a": [0.1, 0.01, 1.5, None], "b": [1, 10, None, None]})
    ln_table = table.eval_expression_list([col("a").ln(), col("b").ln()])
    assert lists_close_with_nones(
        [math.log(v) if v is not None else v for v in table.get_column("a").to_pylist()],
        ln_table.get_column("a").to_pylist(),
    )
    assert lists_close_with_nones(
        [math.log(v) if v is not None else v for v in table.get_column("b").to_pylist()],
        ln_table.get_column("b").to_pylist(),
    )


def test_table_ln_bad_input() -> None:
    table = MicroPartition.from_pydict({"a": ["a", "b", "c"]})

    with pytest.raises(ValueError, match="Expected input to log to be numeric"):
        table.eval_expression_list([col("a").ln()])


def test_table_exp() -> None:
    table = MicroPartition.from_pydict({"a": [0.1, 0.01, None], "b": [1, 10, None]})
    exp_table = table.eval_expression_list([col("a").exp(), col("b").exp()])
    assert lists_close_with_nones(
        [1.1051709180756477, 1.010050167084168, None],
        exp_table.get_column("a").to_pylist(),
    )
    assert lists_close_with_nones(
        [2.718281828459045, 22026.465794806718, None],
        exp_table.get_column("b").to_pylist(),
    )


def test_table_numeric_sqrt() -> None:
    table = MicroPartition.from_pydict({"a": [4, 9, None, 16, 25, None], "b": [2.25, 0.81, None, 1, 10.24, None]})
    sqrt_table = table.eval_expression_list([col("a").sqrt(), col("b").sqrt()])
    assert [math.sqrt(v) if v is not None else v for v in table.get_column("a").to_pylist()] == sqrt_table.get_column(
        "a"
    ).to_pylist()
    assert [math.sqrt(v) if v is not None else v for v in table.get_column("b").to_pylist()] == sqrt_table.get_column(
        "b"
    ).to_pylist()


@pytest.mark.parametrize(
    "left, right",
    [
        pytest.param([1, 2, 3], [4, 5, 6], id="NoNulls"),
        pytest.param([1, 2, None], [4, None, 6], id="WithNulls"),
    ],
)
def test_table_shift_left(left, right) -> None:
    table = MicroPartition.from_pydict({"a": left, "b": right})
    shift_left_table = table.eval_expression_list([col("a") << (col("b"))])
    assert shift_left_table.get_column("a").to_pylist() == [
        i << j if i is not None and j is not None else None for i, j in zip(left, right)
    ]


def test_table_shift_left_with_scalar() -> None:
    table = MicroPartition.from_pydict({"a": [1, 2, 4]})
    shift_left_table = table.eval_expression_list([col("a") << (1)])
    assert [1 << 1, 2 << 1, 4 << 1] == shift_left_table.get_column("a").to_pylist()


def test_table_shift_left_bad_input() -> None:
    table = MicroPartition.from_pydict({"a": ["a", "b", "c"]})

    with pytest.raises(ValueError, match="Cannot operate shift left on types: Utf8, Utf8"):
        table.eval_expression_list([col("a") << (col("a"))])


def test_table_shift_left_bad_shift() -> None:
    table = MicroPartition.from_pydict({"a": [1, 2, 4], "b": [3, 2, 1]})

    with pytest.raises(ValueError, match="Cannot operate shift left on types: Int64, Utf8"):
        table.eval_expression_list([col("a") << (lit("a"))])


def test_table_shift_left_negative_bits() -> None:
    table = MicroPartition.from_pydict({"a": [1, 2, 4], "b": [3, 2, -1]})

    with pytest.raises(ValueError, match="Cannot shift left by a negative number"):
        table.eval_expression_list([col("a") << (col("b"))])


def test_table_shift_left_syntactic_sugar() -> None:
    table = MicroPartition.from_pydict({"a": [1, 2, 4]})
    shift_table = table.eval_expression_list([col("a").shift_left(1)])
    assert [1 << 1, 2 << 1, 4 << 1] == shift_table.get_column("a").to_pylist()


@pytest.mark.parametrize(
    "left, right",
    [
        pytest.param([1, 2, 3], [4, 5, 6], id="NoNulls"),
        pytest.param([1, 2, None], [4, None, 6], id="WithNulls"),
    ],
)
def test_table_shift_right(left, right) -> None:
    table = MicroPartition.from_pydict({"a": left, "b": right})
    shift_right_table = table.eval_expression_list([col("a") >> (col("b"))])
    assert shift_right_table.get_column("a").to_pylist() == [
        i >> j if i is not None and j is not None else None for i, j in zip(left, right)
    ]


def test_table_shift_right_with_scalar() -> None:
    table = MicroPartition.from_pydict({"a": [1, 2, 4]})
    shift_right_table = table.eval_expression_list([col("a") >> (1)])
    assert [1 >> 1, 2 >> 1, 4 >> 1] == shift_right_table.get_column("a").to_pylist()


def test_table_shift_right_bad_input() -> None:
    table = MicroPartition.from_pydict({"a": ["a", "b", "c"]})

    with pytest.raises(ValueError, match="Cannot operate shift right on types: Utf8, Utf8"):
        table.eval_expression_list([col("a") >> (col("a"))])


def test_table_shift_right_bad_shift() -> None:
    table = MicroPartition.from_pydict({"a": [8, 4, 2], "b": [3, 2, 1]})

    with pytest.raises(ValueError, match="Cannot operate shift right on types: Int64, Utf8"):
        table.eval_expression_list([col("a") >> (lit("a"))])


def test_table_shift_right_negative_bits() -> None:
    table = MicroPartition.from_pydict({"a": [8, 4, 2], "b": [3, 2, -1]})

    with pytest.raises(ValueError, match="Cannot shift right by a negative number"):
        table.eval_expression_list([col("a") >> (col("b"))])


def test_table_shift_right_syntactic_sugar() -> None:
    table = MicroPartition.from_pydict({"a": [1, 2, 4]})
    shift_table = table.eval_expression_list([col("a").shift_right(1)])
    assert [1 >> 1, 2 >> 1, 4 >> 1] == shift_table.get_column("a").to_pylist()



import pytest

import daft


@pytest.mark.skip(
    "invoke manually via `uv run tests/sql/test_table_functions/test_read_iceberg.py <metadata_location>`"
)
def test_read_iceberg(metadata_location):
    df = daft.sql(f"SELECT * FROM read_iceberg('{metadata_location}')")
    print(df.collect())


if __name__ == "__main__":
    import sys

    if len(sys.argv) < 2:
        print("usage: test_read_iceberg.py <metadata_location>")
        sys.exit(1)
    test_read_iceberg(metadata_location=sys.argv[1])



import logging

import pytest


def test_logger_initialization():
    import daft

    rust_level = daft.daft.get_max_log_level()

    assert rust_level == "WARN"


def test_debug_logger():
    import daft
    from daft.logging import setup_debug_logger

    setup_debug_logger()
    rust_level = daft.daft.get_max_log_level()
    assert rust_level == "DEBUG"


@pytest.mark.parametrize(
    "level, expected",
    [
        (logging.DEBUG, "DEBUG"),
        (logging.INFO, "INFO"),
        (logging.WARNING, "WARN"),
        (logging.ERROR, "ERROR"),
    ],
)
def test_refresh_logger(level, expected):
    import logging

    import daft

    logging.getLogger().setLevel(level)
    daft.daft.refresh_logger()

    rust_level = daft.daft.get_max_log_level()
    assert rust_level == expected



import pytest

import daft
from tests.benchmarks.conftest import IS_CI

ENGINES = ["native", "python"]


@pytest.mark.skipif(IS_CI, reason="Write benchmarks are not run in CI")
@pytest.mark.benchmark(group="write")
@pytest.mark.parametrize("engine", ENGINES)
@pytest.mark.parametrize(
    "file_type, target_file_size, target_row_group_size",
    [
        ("parquet", None, None),
        (
            "parquet",
            5 * 1024 * 1024,
            1024 * 1024,
        ),  # 5MB target file size, 1MB target row group size
        ("csv", None, None),
    ],
)
@pytest.mark.parametrize("partition_cols", [None, ["L_SHIPMODE"]])
@pytest.mark.parametrize("get_df", ["in-memory"], indirect=True)
def test_streaming_write(
    tmp_path,
    get_df,
    benchmark_with_memray,
    engine,
    file_type,
    target_file_size,
    target_row_group_size,
    partition_cols,
):
    get_df, num_parts = get_df
    daft_df = get_df("lineitem")

    def f():
        if engine == "native":
            daft.context.set_runner_native()
        elif engine == "python":
            daft.context.set_runner_py()
        else:
            raise ValueError(f"{engine} unsupported")

        ctx = daft.context.execution_config_ctx(
            parquet_target_filesize=target_file_size,
            parquet_target_row_group_size=target_row_group_size,
            csv_target_filesize=target_file_size,
        )

        with ctx:
            if file_type == "parquet":
                return daft_df.write_parquet(tmp_path, partition_cols=partition_cols)
            elif file_type == "csv":
                return daft_df.write_csv(tmp_path, partition_cols=partition_cols)
            else:
                raise ValueError(f"{file_type} unsupported")

    benchmark_group = f"parts-{num_parts}-partition-cols-{partition_cols}-file-type-{file_type}-target-file-size-{target_file_size}-target-row-group-size-{target_row_group_size}"
    result_files = benchmark_with_memray(f, benchmark_group).to_pydict()["path"]
    read_back = daft.read_parquet(result_files) if file_type == "parquet" else daft.read_csv(result_files)
    assert read_back.count_rows() == daft_df.count_rows()



from __future__ import annotations

import contextlib
import io
import os
import pathlib
import shutil
from typing import Generator, TypeVar

import numpy as np
import pytest
import s3fs
from PIL import Image

import daft

T = TypeVar("T")

YieldFixture = Generator[T, None, None]


###
# Config fixtures
###


@pytest.fixture(scope="session")
def minio_io_config() -> daft.io.IOConfig:
    return daft.io.IOConfig(
        s3=daft.io.S3Config(
            endpoint_url="http://127.0.0.1:9000",
            key_id="minioadmin",
            access_key="minioadmin",
            use_ssl=False,
        )
    )


@pytest.fixture(scope="session")
def aws_public_s3_config(request) -> daft.io.IOConfig:
    # Use anonymous mode to avoid having to search for credentials in the Github Runner
    # If pytest is run with `--credentials` then we set anonymous=None to go down the credentials chain
    anonymous = None if request.config.getoption("--credentials") else True

    return daft.io.IOConfig(
        s3=daft.io.S3Config(
            # NOTE: no keys or endpoints specified for an AWS public s3 bucket
            region_name="us-west-2",
            anonymous=anonymous,
        )
    )


@pytest.fixture(scope="session")
def gcs_public_config(request) -> daft.io.IOConfig:
    # Use anonymous mode to avoid having to search for credentials in the Github Runner
    # If pytest is run with `--credentials` then we set anonymous=None to go down the credentials chain
    anonymous = None if request.config.getoption("--credentials") else True
    return daft.io.IOConfig(gcs=daft.io.GCSConfig(project_id=None, anonymous=anonymous))


@pytest.fixture(scope="session")
def azure_storage_public_config() -> daft.io.IOConfig:
    return daft.io.IOConfig(
        azure=daft.io.AzureConfig(
            storage_account="dafttestdata",
            anonymous=True,
        )
    )


@pytest.fixture(scope="session")
def nginx_config() -> tuple[str, pathlib.Path]:
    """Returns the (nginx_server_url, static_files_tmpdir) as a tuple."""
    return (
        "http://127.0.0.1:8080",
        pathlib.Path("/tmp/daft-integration-testing/nginx"),
    )


@pytest.fixture(scope="session", params=["standard", "adaptive"], ids=["standard", "adaptive"])
def retry_server_s3_config(request) -> daft.io.IOConfig:
    """Returns the URL to the local retry_server fixture."""
    retry_mode = request.param
    return daft.io.IOConfig(
        s3=daft.io.S3Config(endpoint_url="http://127.0.0.1:8001", anonymous=True, num_tries=10, retry_mode=retry_mode)
    )


###
# Mounting utilities: mount data and perform cleanup at the end of each test
###


@contextlib.contextmanager
def minio_create_bucket(
    minio_io_config: daft.io.IOConfig, bucket_name: str = "my-minio-bucket"
) -> YieldFixture[list[str]]:
    """Creates a bucket in MinIO.

    Yields a s3fs FileSystem
    """
    fs = s3fs.S3FileSystem(
        key=minio_io_config.s3.key_id,
        password=minio_io_config.s3.access_key,
        client_kwargs={"endpoint_url": minio_io_config.s3.endpoint_url},
    )
    if fs.exists(bucket_name):
        fs.rm(bucket_name, recursive=True)
    fs.mkdir(bucket_name)
    try:
        yield fs
    finally:
        fs.rm(bucket_name, recursive=True)


@contextlib.contextmanager
def mount_data_minio(
    minio_io_config: daft.io.IOConfig, folder: pathlib.Path, bucket_name: str = "my-minio-bucket"
) -> YieldFixture[list[str]]:
    """Mounts data in `folder` into files in minio.

    Yields a list of S3 URLs
    """
    with minio_create_bucket(minio_io_config=minio_io_config, bucket_name=bucket_name) as fs:
        urls = []
        for p in folder.glob("**/*"):
            if not p.is_file():
                continue
            key = str(p.relative_to(folder))
            url = f"s3://{bucket_name}/{key}"
            fs.write_bytes(url, p.read_bytes())
            urls.append(url)

        yield urls


@contextlib.contextmanager
def mount_data_nginx(nginx_config: tuple[str, pathlib.Path], folder: pathlib.Path) -> YieldFixture[list[str]]:
    """Mounts data in `folder` into servable static files in NGINX.

    Yields a list of HTTP URLs
    """
    server_url, static_assets_tmpdir = nginx_config

    # Cleanup any old stuff in mount folder
    for item in os.listdir(static_assets_tmpdir):
        path = static_assets_tmpdir / item
        if path.is_dir():
            shutil.rmtree(path)
        else:
            os.remove(path)

    # Copy data to mount folder
    for item in os.listdir(folder):
        src = folder / item
        dest = static_assets_tmpdir / item
        if src.is_dir():
            shutil.copytree(str(src), str(dest))
        else:
            shutil.copy2(src, dest)

    try:
        yield [f"{server_url}/{p.relative_to(folder)}" for p in folder.glob("**/*") if p.is_file()]
    finally:
        for item in os.listdir(static_assets_tmpdir):
            path = static_assets_tmpdir / item
            if path.is_dir():
                shutil.rmtree(static_assets_tmpdir / item)
            else:
                os.remove(static_assets_tmpdir / item)


###
# Image data test fixtures
###


@pytest.fixture(scope="session")
def image_data() -> YieldFixture[bytes]:
    """Bytes of a small image."""
    bio = io.BytesIO()
    image = Image.fromarray(np.ones((3, 3)).astype(np.uint8))
    image.save(bio, format="JPEG")
    return bio.getvalue()


@pytest.fixture(scope="function")
def image_data_folder(image_data, tmpdir) -> YieldFixture[str]:
    """Dumps 10 small JPEG files into a tmpdir."""
    tmpdir = pathlib.Path(tmpdir)

    for i in range(10):
        fp = tmpdir / f"{i}.jpeg"
        fp.write_bytes(image_data)

    yield tmpdir


@pytest.fixture(scope="function")
def mock_http_image_urls(
    nginx_config: tuple[str, pathlib.Path], image_data_folder: pathlib.Path
) -> YieldFixture[list[str]]:
    """Uses the docker-compose Nginx server to serve HTTP image URLs.

    This fixture yields:
        list[str]: URLs of files available on the HTTP server
    """
    with mount_data_nginx(nginx_config, image_data_folder) as urls:
        yield urls


@pytest.fixture(scope="function")
def minio_image_data_fixture(minio_io_config, image_data_folder) -> YieldFixture[list[str]]:
    """Populates the minio session with some fake data and yields (S3Config, paths)."""
    with mount_data_minio(minio_io_config, image_data_folder) as urls:
        yield urls


@pytest.fixture(scope="session")
def small_images_s3_paths() -> list[str]:
    """Paths to small *.jpg files in a public S3 bucket."""
    return [f"s3://daft-public-data/test_fixtures/small_images/rickroll{i}.jpg" for i in range(6)] + [
        f"s3a://daft-public-data/test_fixtures/small_images/rickroll{i}.jpg" for i in range(6)
    ]



from __future__ import annotations

import contextlib
import json
import os
import pathlib
import tempfile
from typing import Any

import pytest

import daft
from daft.datatype import DataType
from daft.logical.schema import Schema
from daft.recordbatch import MicroPartition, recordbatch_io
from daft.runners.partitioning import TableReadOptions


def test_read_input(tmpdir):
    tmpdir = pathlib.Path(tmpdir)
    data = {"foo": [1, 2, 3]}
    with open(tmpdir / "file.json", "w") as f:
        for i in range(1, 4):
            json.dump({"foo": i}, f)
            f.write("\n")

    schema = Schema._from_field_name_and_types([("foo", DataType.int64())])

    # Test pathlib, str and IO
    assert recordbatch_io.read_json(tmpdir / "file.json", schema=schema).to_pydict() == data
    assert recordbatch_io.read_json(str(tmpdir / "file.json"), schema=schema).to_pydict() == data


@contextlib.contextmanager
def _json_write_helper(data: dict[str, list[Any]]):
    with tempfile.TemporaryDirectory() as directory_name:
        first_key = next(iter(data.keys()))
        data_len = len(data[first_key])
        for k in data:
            assert len(data[k]) == data_len
        file = os.path.join(directory_name, "tempfile")
        with open(file, "w", newline="") as f:
            for i in range(data_len):
                f.write(json.dumps({k: data[k][i] for k in data}) + "\n")
        yield file


@pytest.mark.parametrize(
    ["data", "expected_dtype"],
    [
        ("1", DataType.string()),
        ("foo", DataType.string()),
        ("1.5", DataType.string()),
        ("True", DataType.string()),
        ("None", DataType.string()),
        (1, DataType.int64()),
        (1.5, DataType.float64()),
        (True, DataType.bool()),
        (None, DataType.null()),
        ({"foo": 1}, DataType.struct({"foo": DataType.int64()})),
        ([1, None, 2], DataType.list(DataType.int64())),
    ],
)
def test_json_infer_schema(data, expected_dtype):
    with _json_write_helper(
        {
            "id": [1, 2, 3],
            "data": [data, data, None],
        }
    ) as f:
        schema = Schema.from_json(f)
        assert schema == Schema._from_field_name_and_types([("id", DataType.int64()), ("data", expected_dtype)])


@pytest.mark.parametrize(
    ["data", "expected_data_series"],
    [
        [1, daft.Series.from_pylist([1, 1, None])],
        ["foo", daft.Series.from_pylist(["foo", "foo", None])],
        [1.5, daft.Series.from_pylist([1.5, 1.5, None])],
        (True, daft.Series.from_pylist([True, True, None])),
        ([[], [1, 2], None], daft.Series.from_pylist([[[], [1, 2], None], [[], [1, 2], None], None])),
        ({"foo": 1}, daft.Series.from_pylist([{"foo": 1}, {"foo": 1}, None])),
    ],
)
def test_json_read_data(data, expected_data_series):
    with _json_write_helper(
        {
            "id": [1, 2, 3],
            "data": [data, data, None],
        }
    ) as f:
        schema = Schema._from_field_name_and_types(
            [("id", DataType.int64()), ("data", expected_data_series.datatype())]
        )
        expected = MicroPartition.from_pydict(
            {
                "id": [1, 2, 3],
                "data": expected_data_series,
            }
        )
        table = recordbatch_io.read_json(f, schema)
        assert table.to_arrow() == expected.to_arrow(), f"Expected:\n{expected}\n\nReceived:\n{table}"


def test_json_read_data_limit_rows():
    with _json_write_helper(
        {
            "id": [1, 2, 3],
            "data": [1, 2, None],
        }
    ) as f:
        schema = Schema._from_field_name_and_types([("id", DataType.int64()), ("data", DataType.int64())])
        expected = MicroPartition.from_pydict(
            {
                "id": [1, 2],
                "data": [1, 2],
            }
        )
        table = recordbatch_io.read_json(f, schema, read_options=TableReadOptions(num_rows=2))
        assert table.to_arrow() == expected.to_arrow(), f"Expected:\n{expected}\n\nReceived:\n{table}"


def test_json_read_data_select_columns():
    with _json_write_helper(
        {
            "id": [1, 2, 3],
            "data": [1, 2, None],
        }
    ) as f:
        schema = Schema._from_field_name_and_types([("id", DataType.int64()), ("data", DataType.int64())])
        expected = MicroPartition.from_pydict(
            {
                "data": [1, 2, None],
            }
        )
        table = recordbatch_io.read_json(f, schema, read_options=TableReadOptions(column_names=["data"]))
        assert table.to_arrow() == expected.to_arrow(), f"Expected:\n{expected}\n\nReceived:\n{table}"



from daft.daft import ImageFormat, ImageMode, PySeries

def decode(s: PySeries, raise_error_on_failure: bool, mode: ImageMode | None = None) -> PySeries: ...
def encode(s: PySeries, image_format: ImageFormat) -> PySeries: ...
def resize(s: PySeries, w: int, h: int) -> PySeries: ...
def to_mode(s: PySeries, mode: ImageMode) -> PySeries: ...



import os
import platform
import urllib.parse
import urllib.request

from daft import get_build_type, get_version


def scarf_telemetry(runner: str):
    # Track analytics for Daft usage via Scarf.

    version = get_version()
    build_type = get_build_type()
    scarf_opt_out = os.getenv("SCARF_NO_ANALYTICS") == "true" or os.getenv("DO_NOT_TRACK") == "true"
    daft_analytics_enabled = os.getenv("DAFT_ANALYTICS_ENABLED") != "0"

    # Skip analytics for dev builds or if user opted out
    if build_type == "dev" or scarf_opt_out or daft_analytics_enabled:
        return None, None

    try:
        python_version = ".".join(platform.python_version().split(".")[:2])

        params = {
            "version": version,
            "platform": platform.system(),
            "python": python_version,
            "arch": platform.machine(),
            "runner": runner,
        }

        # Prepare the query string
        query_string = urllib.parse.urlencode(params)

        # Make the GET request
        url = f"https://daft.gateway.scarf.sh/daft-runner?{query_string}"
        with urllib.request.urlopen(url) as response:
            return f"Response status: {response.status}", runner

    except Exception as e:
        return f"Analytics error: {e!s}", None



from __future__ import annotations

from typing import Generator, Iterator, TypeVar

import pyarrow as pa
import pytest

import daft
import daft.catalog

pyiceberg = pytest.importorskip("pyiceberg")

PYARROW_LOWER_BOUND_SKIP = tuple(int(s) for s in pa.__version__.split(".") if s.isnumeric()) < (9, 0, 0)
pytestmark = pytest.mark.skipif(
    PYARROW_LOWER_BOUND_SKIP, reason="iceberg writes not supported on old versions of pyarrow"
)

import tenacity
from pyiceberg.catalog import Catalog, load_catalog
from pyiceberg.table import Table

T = TypeVar("T")

YieldFixture = Generator[T, None, None]

local_tables_names = [
    "test_all_types",
    "test_limit",
    "test_null_nan",
    "test_null_nan_rewritten",
    "test_partitioned_by_bucket",
    "test_partitioned_by_days",
    "test_partitioned_by_hours",
    "test_partitioned_by_identity",
    "test_partitioned_by_months",
    "test_partitioned_by_truncate",
    "test_partitioned_by_years",
    "test_positional_mor_deletes",
    "test_positional_mor_double_deletes",
    "test_table_sanitized_character",
    "test_table_version",
    "test_uuid_and_fixed_unpartitioned",
]


cloud_tables_names = [
    "azure.test",
    # TODO(Kevin): Add more tables from more cloud providers
]


@pytest.fixture(scope="session")
def local_iceberg_catalog() -> Iterator[tuple[str, Catalog]]:
    cat = load_catalog(
        "local",
        **{
            "type": "rest",
            "uri": "http://localhost:8181",
            "s3.endpoint": "http://localhost:9000",
            "s3.access-key-id": "admin",
            "s3.secret-access-key": "password",
        },
    )

    # ensure all tables are available
    for name in local_tables_names:
        _load_table(cat, name)

    catalog_name = "_local_iceberg_catalog"
    daft.catalog.register_python_catalog(cat, name=catalog_name)
    yield catalog_name, cat
    daft.catalog.unregister_catalog(catalog_name=catalog_name)


@pytest.fixture(scope="session")
def azure_iceberg_catalog() -> Iterator[tuple[str, Catalog]]:
    cat = load_catalog(
        "default",
        **{
            "uri": "sqlite:///tests/assets/pyiceberg_catalog.db",
            "adlfs.account-name": "dafttestdata",
        },
    )

    catalog_name = "_azure_iceberg_catalog"
    daft.catalog.register_python_catalog(cat, name=catalog_name)
    yield catalog_name, cat
    daft.catalog.unregister_catalog(catalog_name=catalog_name)


@tenacity.retry(
    stop=tenacity.stop_after_delay(60),
    retry=tenacity.retry_if_exception_type(pyiceberg.exceptions.NoSuchTableError),
    wait=tenacity.wait_fixed(5),
    reraise=True,
)
def _load_table(catalog, name) -> Table:
    return catalog.load_table(f"default.{name}")


@pytest.fixture(scope="function", params=local_tables_names)
def local_iceberg_tables(request, local_iceberg_catalog) -> Iterator[str]:
    NAMESPACE = "default"
    table_name = request.param
    yield f"{NAMESPACE}.{table_name}"


@pytest.fixture(scope="function", params=cloud_tables_names)
def azure_iceberg_table(request, azure_iceberg_catalog) -> Iterator[str]:
    yield request.param



# isort: dont-add-import: from __future__ import annotations


from typing import TYPE_CHECKING, Callable, Dict, Optional, Union

from daft import context, from_pydict
from daft.api_annotations import PublicAPI
from daft.daft import ScanOperatorHandle, StorageConfig
from daft.dataframe import DataFrame
from daft.datatype import DataType
from daft.logical.builder import LogicalPlanBuilder
from daft.sql.sql_connection import SQLConnection
from daft.sql.sql_scan import PartitionBoundStrategy, SQLScanOperator

if TYPE_CHECKING:
    from sqlalchemy.engine import Connection


@PublicAPI
def read_sql(
    sql: str,
    conn: Union[Callable[[], "Connection"], str],
    partition_col: Optional[str] = None,
    num_partitions: Optional[int] = None,
    partition_bound_strategy: str = "min-max",
    disable_pushdowns_to_sql: bool = False,
    infer_schema: bool = True,
    infer_schema_length: int = 10,
    schema: Optional[Dict[str, DataType]] = None,
) -> DataFrame:
    """Create a DataFrame from the results of a SQL query.

    Args:
        sql (str): SQL query to execute
        conn (Union[Callable[[], Connection], str]): SQLAlchemy connection factory or database URL
        partition_col (Optional[str]): Column to partition the data by, defaults to None
        num_partitions (Optional[int]): Number of partitions to read the data into,
            defaults to None, which will lets Daft determine the number of partitions.
            If specified, `partition_col` must also be specified.
        partition_bound_strategy (str): Strategy to determine partition bounds, either "min-max" or "percentile", defaults to "min-max"
        disable_pushdowns_to_sql (bool): Whether to disable pushdowns to the SQL query, defaults to False
        infer_schema (bool): Whether to turn on schema inference, defaults to True. If set to False, the schema parameter must be provided.
        infer_schema_length (int): The number of rows to scan when inferring the schema, defaults to 10. If infer_schema is False, this parameter is ignored. Note that if Daft is able to use ConnectorX to infer the schema, this parameter is ignored as ConnectorX is an Arrow backed driver.
        schema (Optional[Dict[str, DataType]]): A mapping of column names to datatypes. If infer_schema is False, this schema is used as the definitive schema for the data, otherwise it is used as a schema hint that is applied after the schema is inferred.
            This can be useful if the types can be more precisely determined than what the inference can provide (e.g., if a column can be declared as a fixed-sized list rather than a list).

    Returns:
        DataFrame: Dataframe containing the results of the query

    .. NOTE::
        #. Supported dialects:
            Daft uses `SQLGlot <https://sqlglot.com/sqlglot.html>`_ to build and translate SQL queries between dialects. For a list of supported dialects, see `SQLGlot's dialect documentation <https://sqlglot.com/sqlglot/dialects.html>`_.

        #. Partitioning:
            When `partition_col` is specified, the function partitions the query based on that column.
            You can define `num_partitions` or leave it to Daft to decide.
            Daft uses the `partition_bound_strategy` parameter to determine the partitioning strategy:
            - `min_max`: Daft calculates the minimum and maximum values of the specified column, then partitions the query using equal ranges between the minimum and maximum values.
            - `percentile`: Daft calculates the specified column's percentiles via a `PERCENTILE_DISC` function to determine partitions (e.g., for `num_partitions=3`, it uses the 33rd and 66th percentiles).

        #. Execution:
            Daft executes SQL queries using using `ConnectorX <https://sfu-db.github.io/connector-x/intro.html>`_ or `SQLAlchemy <https://docs.sqlalchemy.org/en/20/orm/quickstart.html#create-an-engine>`_,
            preferring ConnectorX unless a SQLAlchemy connection factory is specified or the database dialect is unsupported by ConnectorX.

        #. Pushdowns:
            Daft pushes down operations such as filtering, projections, and limits into the SQL query when possible.
            You can disable pushdowns by setting `disable_pushdowns_to_sql=True`, which will execute the SQL query as is.

    Example:
        Read data from a SQL query and a database URL:

        >>> df = daft.read_sql("SELECT * FROM my_table", "sqlite:///my_database.db")

        Read data from a SQL query and a SQLAlchemy connection factory:

        >>> def create_conn():
        ...     return sqlalchemy.create_engine("sqlite:///my_database.db").connect()
        >>> df = daft.read_sql("SELECT * FROM my_table", create_conn)

        Read data from a SQL query and partition the data by a column:

        >>> df = daft.read_sql("SELECT * FROM my_table", "sqlite:///my_database.db", partition_col="id")

        Read data from a SQL query and partition the data into 3 partitions:

        >>> df = daft.read_sql("SELECT * FROM my_table", "sqlite:///my_database.db", partition_col="id", num_partitions=3)
    """
    if num_partitions is not None and partition_col is None:
        raise ValueError("Failed to execute sql: partition_col must be specified when num_partitions is specified")

    if not infer_schema and schema is None:
        raise ValueError(
            "Cannot read DataFrame with infer_schema=False and schema=None, please provide a schema or set infer_schema=True"
        )

    io_config = context.get_context().daft_planning_config.default_io_config
    storage_config = StorageConfig(True, io_config)

    sql_conn = SQLConnection.from_url(conn) if isinstance(conn, str) else SQLConnection.from_connection_factory(conn)
    sql_operator = SQLScanOperator(
        sql,
        sql_conn,
        storage_config,
        disable_pushdowns_to_sql,
        infer_schema,
        infer_schema_length,
        schema,
        partition_col=partition_col,
        num_partitions=num_partitions,
        partition_bound_strategy=PartitionBoundStrategy.from_str(partition_bound_strategy),
    )
    handle = ScanOperatorHandle.from_python_scan_operator(sql_operator)
    builder = LogicalPlanBuilder.from_tabular_scan(scan_operator=handle)

    return DataFrame(builder)


@PublicAPI
def sql(sql: str) -> DataFrame:
    """Create a DataFrame from an SQL query.

    EXPERIMENTAL: This features is early in development and will change.

    Args:
        sql (str): SQL query to execute

    Returns:
        DataFrame: Dataframe containing the results of the query
    """
    return from_pydict({})



from __future__ import annotations

import datetime
import math

import pandas as pd
import pytest
import sqlalchemy

import daft
from tests.conftest import assert_df_equals
from tests.integration.sql.conftest import EMPTY_TEST_TABLE_NAME, TEST_TABLE_NAME


@pytest.fixture(scope="session")
def pdf(test_db):
    return pd.read_sql_query(f"SELECT * FROM {TEST_TABLE_NAME}", test_db)


@pytest.mark.integration()
def test_sql_show(test_db) -> None:
    df = daft.read_sql(f"SELECT * FROM {TEST_TABLE_NAME}", test_db)
    df.show()


@pytest.mark.integration()
def test_sql_create_dataframe_ok(test_db, pdf) -> None:
    df = daft.read_sql(f"SELECT * FROM {TEST_TABLE_NAME}", test_db)

    assert_df_equals(df.to_pandas(coerce_temporal_nanoseconds=True), pdf, sort_key="id")


@pytest.mark.integration()
@pytest.mark.parametrize("num_partitions", [2, 3, 4])
@pytest.mark.parametrize("partition_bound_strategy", ["min-max", "percentile"])
def test_sql_partitioned_read(test_db, num_partitions, partition_bound_strategy, pdf) -> None:
    row_size_bytes = daft.from_pandas(pdf).schema().estimate_row_size_bytes()
    num_rows_per_partition = len(pdf) / num_partitions
    with daft.execution_config_ctx(
        read_sql_partition_size_bytes=math.ceil(row_size_bytes * num_rows_per_partition),
        scan_tasks_min_size_bytes=0,
        scan_tasks_max_size_bytes=0,
    ):
        df = daft.read_sql(
            f"SELECT * FROM {TEST_TABLE_NAME}",
            test_db,
            partition_col="id",
            partition_bound_strategy=partition_bound_strategy,
        )
        assert df.num_partitions() == num_partitions
        assert_df_equals(df.to_pandas(coerce_temporal_nanoseconds=True), pdf, sort_key="id")


@pytest.mark.integration()
@pytest.mark.parametrize("num_partitions", [1, 2, 3, 4])
@pytest.mark.parametrize("partition_col", ["id", "float_col", "date_col", "date_time_col"])
@pytest.mark.parametrize("partition_bound_strategy", ["min-max", "percentile"])
def test_sql_partitioned_read_with_custom_num_partitions_and_partition_col(
    test_db, num_partitions, partition_col, partition_bound_strategy, pdf
) -> None:
    with daft.execution_config_ctx(
        scan_tasks_min_size_bytes=0,
        scan_tasks_max_size_bytes=0,
    ):
        df = daft.read_sql(
            f"SELECT * FROM {TEST_TABLE_NAME}",
            test_db,
            partition_col=partition_col,
            num_partitions=num_partitions,
            partition_bound_strategy=partition_bound_strategy,
        )
        assert df.num_partitions() == num_partitions
        assert_df_equals(df.to_pandas(coerce_temporal_nanoseconds=True), pdf, sort_key="id")


@pytest.mark.integration()
@pytest.mark.parametrize("num_partitions", [0, 1, 2])
@pytest.mark.parametrize("partition_col", ["id", "string_col"])
def test_sql_partitioned_read_on_empty_table(empty_test_db, num_partitions, partition_col) -> None:
    with daft.execution_config_ctx(
        scan_tasks_min_size_bytes=0,
        scan_tasks_max_size_bytes=0,
    ):
        df = daft.read_sql(
            f"SELECT * FROM {EMPTY_TEST_TABLE_NAME}",
            empty_test_db,
            partition_col=partition_col,
            num_partitions=num_partitions,
            schema={"id": daft.DataType.int64(), "string_col": daft.DataType.string()},
        )
        assert df.num_partitions() == 1
        empty_pdf = pd.read_sql_query(
            f"SELECT * FROM {EMPTY_TEST_TABLE_NAME}", empty_test_db, dtype={"id": "int64", "string_col": "str"}
        )
        assert_df_equals(df.to_pandas(), empty_pdf, sort_key="id")


@pytest.mark.integration()
@pytest.mark.parametrize("num_partitions", [1, 2, 3, 4])
def test_sql_partitioned_read_with_non_uniformly_distributed_column(test_db, num_partitions, pdf) -> None:
    with daft.execution_config_ctx(
        scan_tasks_min_size_bytes=0,
        scan_tasks_max_size_bytes=0,
    ):
        df = daft.read_sql(
            f"SELECT * FROM {TEST_TABLE_NAME}",
            test_db,
            partition_col="non_uniformly_distributed_col",
            num_partitions=num_partitions,
        )
        assert df.num_partitions() == num_partitions
        assert_df_equals(df.to_pandas(coerce_temporal_nanoseconds=True), pdf, sort_key="id")


@pytest.mark.integration()
@pytest.mark.parametrize("partition_col", ["string_col", "null_col"])
def test_sql_partitioned_read_with_non_partionable_column(test_db, partition_col) -> None:
    with pytest.raises(ValueError, match="Failed to get partition bounds"):
        df = daft.read_sql(
            f"SELECT * FROM {TEST_TABLE_NAME}",
            test_db,
            partition_col=partition_col,
            num_partitions=2,
        )
        df = df.collect()


@pytest.mark.integration()
def test_sql_read_with_partition_num_without_partition_col(test_db) -> None:
    with pytest.raises(ValueError, match="Failed to execute sql"):
        daft.read_sql(f"SELECT * FROM {TEST_TABLE_NAME}", test_db, num_partitions=2)


@pytest.mark.integration()
@pytest.mark.parametrize(
    "operator",
    ["<", ">", "=", "!=", ">=", "<="],
)
@pytest.mark.parametrize(
    "column, value",
    [
        ("id", 100),
        ("float_col", 100.0123),
        ("string_col", "row_100"),
        ("bool_col", True),
        ("date_col", datetime.date(2021, 1, 1)),
        ("date_time_col", datetime.datetime(2020, 1, 1, 10, 0, 0)),
    ],
)
@pytest.mark.parametrize("num_partitions", [1, 2])
def test_sql_read_with_binary_filter_pushdowns(test_db, column, operator, value, num_partitions, pdf) -> None:
    # Skip invalid comparisons for bool_col
    if column == "bool_col" and operator not in ("=", "!="):
        pytest.skip(f"Operator {operator} not valid for bool_col")

    df = daft.read_sql(
        f"SELECT * FROM {TEST_TABLE_NAME}",
        test_db,
        partition_col="id",
        num_partitions=num_partitions,
    )

    if operator == ">":
        df = df.where(df[column] > value)
        pdf = pdf[pdf[column] > value]
    elif operator == "<":
        df = df.where(df[column] < value)
        pdf = pdf[pdf[column] < value]
    elif operator == "=":
        df = df.where(df[column] == value)
        pdf = pdf[pdf[column] == value]
    elif operator == "!=":
        df = df.where(df[column] != value)
        pdf = pdf[pdf[column] != value]
    elif operator == ">=":
        df = df.where(df[column] >= value)
        pdf = pdf[pdf[column] >= value]
    elif operator == "<=":
        df = df.where(df[column] <= value)
        pdf = pdf[pdf[column] <= value]

    assert_df_equals(df.to_pandas(coerce_temporal_nanoseconds=True), pdf, sort_key="id")


@pytest.mark.integration()
@pytest.mark.parametrize("num_partitions", [1, 2])
def test_sql_read_with_is_null_filter_pushdowns(test_db, num_partitions, pdf) -> None:
    df = daft.read_sql(
        f"SELECT * FROM {TEST_TABLE_NAME}",
        test_db,
        partition_col="id",
        num_partitions=num_partitions,
    )
    df = df.where(df["null_col"].is_null())

    pdf = pdf[pdf["null_col"].isnull()]

    assert_df_equals(df.to_pandas(coerce_temporal_nanoseconds=True), pdf, sort_key="id")


@pytest.mark.integration()
@pytest.mark.parametrize("num_partitions", [1, 2])
def test_sql_read_with_not_null_filter_pushdowns(test_db, num_partitions, pdf) -> None:
    df = daft.read_sql(
        f"SELECT * FROM {TEST_TABLE_NAME}",
        test_db,
        partition_col="id",
        num_partitions=num_partitions,
    )
    df = df.where(df["null_col"].not_null())

    pdf = pdf[pdf["null_col"].notnull()]

    assert_df_equals(df.to_pandas(coerce_temporal_nanoseconds=True), pdf, sort_key="id")


@pytest.mark.integration()
@pytest.mark.parametrize("num_partitions", [1, 2])
def test_sql_read_with_non_pushdowned_predicate(test_db, num_partitions, pdf) -> None:
    df = daft.read_sql(
        f"SELECT * FROM {TEST_TABLE_NAME}",
        test_db,
        partition_col="id",
        num_partitions=num_partitions,
    )

    # If_else is not supported as a pushdown to read_sql, but it should still work
    df = df.where((df["id"] > 100).if_else(df["float_col"] > 150, df["float_col"] < 50))

    pdf = pdf[(pdf["id"] > 100) & (pdf["float_col"] > 150) | (pdf["float_col"] < 50)]

    assert_df_equals(df.to_pandas(coerce_temporal_nanoseconds=True), pdf, sort_key="id")


@pytest.mark.integration()
@pytest.mark.parametrize("num_partitions", [1, 2])
def test_sql_read_with_is_in_filter_pushdown(test_db, num_partitions, pdf) -> None:
    df = daft.read_sql(
        f"SELECT * FROM {TEST_TABLE_NAME}",
        test_db,
        partition_col="id",
        num_partitions=num_partitions,
    )
    df = df.where(df["id"].is_in([1, 2, 3]))

    pdf = pdf[pdf["id"].isin([1, 2, 3])]
    assert_df_equals(df.to_pandas(coerce_temporal_nanoseconds=True), pdf, sort_key="id")


@pytest.mark.integration()
@pytest.mark.parametrize("num_partitions", [1, 2])
def test_sql_read_with_all_pushdowns(test_db, num_partitions) -> None:
    df = daft.read_sql(
        f"SELECT * FROM {TEST_TABLE_NAME}",
        test_db,
        partition_col="id",
        num_partitions=num_partitions,
    )
    df = df.where(~(df["id"] < 1))
    df = df.where(df["string_col"].is_in([f"row_{i}" for i in range(10)]))
    df = df.select(df["id"], df["float_col"], df["string_col"])
    df = df.limit(5)

    df = df.collect()
    assert len(df) == 5
    assert df.column_names == ["id", "float_col", "string_col"]

    pydict = df.to_pydict()
    assert all(i >= 1 for i in pydict["id"])
    assert all(s in [f"row_{i}" for i in range(10)] for s in pydict["string_col"])


@pytest.mark.integration()
@pytest.mark.parametrize("limit", [0, 1, 10, 100, 200])
@pytest.mark.parametrize("num_partitions", [1, 2])
def test_sql_read_with_limit_pushdown(test_db, limit, num_partitions) -> None:
    df = daft.read_sql(
        f"SELECT * FROM {TEST_TABLE_NAME}",
        test_db,
        partition_col="id",
        num_partitions=num_partitions,
    )
    df = df.limit(limit)

    df = df.collect()
    assert len(df) == limit


@pytest.mark.integration()
@pytest.mark.parametrize("num_partitions", [1, 2])
def test_sql_read_with_projection_pushdown(test_db, generated_data, num_partitions) -> None:
    df = daft.read_sql(
        f"SELECT * FROM {TEST_TABLE_NAME}",
        test_db,
        partition_col="id",
        num_partitions=num_partitions,
    )
    df = df.select(df["id"], df["string_col"])

    df = df.collect()
    assert df.column_names == ["id", "string_col"]
    assert len(df) == len(generated_data)


@pytest.mark.integration()
def test_sql_read_without_schema_inference(test_db, generated_data) -> None:
    schema = {
        "id": daft.DataType.int32(),
        "float_col": daft.DataType.float64(),
        "string_col": daft.DataType.string(),
        "bool_col": daft.DataType.bool(),
        "date_col": daft.DataType.date(),
        "date_time_col": daft.DataType.timestamp(timeunit="ns"),
        "null_col": daft.DataType.null(),
        "non_uniformly_distributed_col": daft.DataType.int32(),
    }
    df = daft.read_sql(f"SELECT * FROM {TEST_TABLE_NAME}", test_db, infer_schema=False, schema=schema)
    df = df.collect()

    assert df.column_names == list(schema.keys())
    for field in df.schema():
        assert field.dtype == schema[field.name]

    assert len(df) == len(generated_data)


@pytest.mark.integration()
def test_sql_read_with_schema_inference_and_provided_schema(test_db, pdf) -> None:
    schema = {
        "id": daft.DataType.string(),  # Should be inferred as int32 but overridden to string
    }
    df = daft.read_sql(f"SELECT * FROM {TEST_TABLE_NAME}", test_db, infer_schema=True, schema=schema)
    df = df.collect()

    assert df.schema()["id"].dtype == daft.DataType.string()
    pdf = pdf.astype({"id": str})
    assert_df_equals(df.to_pandas(coerce_temporal_nanoseconds=True), pdf, sort_key="id")


@pytest.mark.integration()
def test_sql_read_with_custom_infer_schema_length(test_db, pdf) -> None:
    # Use sql alchemy to test inference length because connectorx does not use inference length
    def create_conn():
        return sqlalchemy.create_engine(test_db).connect()

    # The null column's first value is None, so the inferred schema should be null
    df = daft.read_sql(
        f"SELECT * FROM {TEST_TABLE_NAME}",
        create_conn,
        infer_schema=True,
        infer_schema_length=1,
    )
    df = df.collect()

    assert df.schema()["null_col"].dtype == daft.DataType.null()
    assert len(df) == len(pdf)

    df = daft.read_sql(
        f"SELECT * FROM {TEST_TABLE_NAME}",
        create_conn,
        infer_schema=True,
        infer_schema_length=2,
    )
    df = df.collect()
    assert df.schema()["null_col"].dtype == daft.DataType.string()
    assert len(df) == len(pdf)
    assert_df_equals(df.to_pandas(coerce_temporal_nanoseconds=True), pdf, sort_key="id")


@pytest.mark.integration()
def test_sql_with_infer_schema_false_and_no_schema_provided(test_db) -> None:
    with pytest.raises(
        ValueError,
        match="Cannot read DataFrame with infer_schema=False and schema=None, please provide a schema or set infer_schema=True",
    ):
        daft.read_sql(f"SELECT * FROM {TEST_TABLE_NAME}", test_db, infer_schema=False)


@pytest.mark.integration()
def test_sql_bad_url() -> None:
    with pytest.raises(RuntimeError, match="Failed to execute sql"):
        daft.read_sql(f"SELECT * FROM {TEST_TABLE_NAME}", "bad_url://")


@pytest.mark.integration()
def test_sql_connection_factory_ok(test_db, pdf) -> None:
    def create_conn():
        return sqlalchemy.create_engine(test_db).connect()

    df = daft.read_sql(f"SELECT * FROM {TEST_TABLE_NAME}", create_conn)
    assert_df_equals(df.to_pandas(coerce_temporal_nanoseconds=True), pdf, sort_key="id")


@pytest.mark.integration()
def test_sql_bad_connection_factory() -> None:
    with pytest.raises(ValueError):
        daft.read_sql(f"SELECT * FROM {TEST_TABLE_NAME}", lambda: "bad_conn")


@pytest.mark.integration()
def test_sql_unsupported_dialect() -> None:
    with pytest.raises(ValueError, match="Unsupported dialect"):
        daft.read_sql(
            f"SELECT * FROM {TEST_TABLE_NAME}",
            "sqlheavy://user:password@localhost:5432/db",
        )



import logging
from collections import defaultdict
from typing import Dict

import ray.experimental  # noqa: TID253

from daft.daft import ResourceRequest
from daft.execution import execution_step
from daft.execution.execution_step import (
    PartitionTaskBuilder,
    SingleOutputPartitionTask,
)
from daft.execution.physical_plan import InProgressPhysicalPlan, stage_id_counter
from daft.runners.partitioning import (
    PartitionT,
)

logger = logging.getLogger(__name__)


def pre_shuffle_merge(
    map_plan: InProgressPhysicalPlan[PartitionT],
    pre_shuffle_merge_threshold: int,
) -> InProgressPhysicalPlan[PartitionT]:
    """Merges intermediate partitions from the map_plan based on memory constraints.

    The function processes incoming map tasks and merges their outputs when:
    1. There are multiple materialized maps available
    2. All remaining maps are materialized and no more input is expected

    Args:
        map_plan: Iterator of partition tasks to be merged

    Yields:
        Merged partition tasks or processed child steps
    """
    NUM_MAPS_THRESHOLD = 4

    stage_id = next(stage_id_counter)
    in_flight_maps: Dict[str, SingleOutputPartitionTask[PartitionT]] = {}
    no_more_input = False

    while True:
        # Get materialized maps by size
        materialized_maps = sorted(
            [
                (
                    p,
                    p.partition_metadata().size_bytes or pre_shuffle_merge_threshold + 1,
                )  # Assume large size if unknown
                for p in in_flight_maps.values()
                if p.done()
            ],
            key=lambda x: x[1],
        )

        # Try to merge materialized maps if we have enough (default to 4) or if we're done with input
        enough_maps = len(materialized_maps) > NUM_MAPS_THRESHOLD
        done_with_input = no_more_input and len(materialized_maps) == len(in_flight_maps)

        if enough_maps or done_with_input:
            # Get location information for all materialized partitions
            partitions = [m[0].result().partition() for m in materialized_maps]
            location_map = ray.experimental.get_object_locations(partitions)

            # Group partitions by node
            node_groups = defaultdict(list)
            unknown_location_group = []  # Special group for partitions without known location

            for partition, size in materialized_maps:
                partition_ref = partition.partition()
                location_info = location_map.get(partition_ref, {})

                if not location_info or "node_ids" not in location_info or not location_info["node_ids"]:
                    unknown_location_group.append((partition, size))
                else:
                    # TODO: Handle multiple locations, with a strategy to select more optimal nodes, e.g. based on memory
                    node_id = location_info["node_ids"][0]  # Use first node if multiple locations exist
                    node_groups[node_id].append((partition, size))

            # Function to create merge groups for a list of partitions
            def create_merge_groups(partitions_list):
                if not partitions_list:
                    return []

                groups = []
                current_group = [partitions_list[0][0]]
                current_size = partitions_list[0][1]

                for partition, size in partitions_list[1:]:
                    if current_size + size > pre_shuffle_merge_threshold:
                        groups.append(current_group)
                        current_group = [partition]
                        current_size = size
                    else:
                        current_group.append(partition)
                        current_size += size

                # Add the last group if it exists and is either:
                # 1. Contains more than 1 partition
                # 2. Is the last group and we're done with input
                # 3. The partition exceeds the memory threshold
                should_add_last_group = (
                    len(current_group) > 1 or done_with_input or current_size > pre_shuffle_merge_threshold
                )
                if current_group and should_add_last_group:
                    groups.append(current_group)

                return groups

            # Process each node's partitions and unknown location partitions
            merge_groups = {}

            # Process node-specific groups
            for node_id, node_partitions in node_groups.items():
                merge_groups[node_id] = create_merge_groups(node_partitions)

            # Process unknown location group
            merge_groups[None] = create_merge_groups(unknown_location_group)

            # Create merge steps and remove processed maps
            for node_id, groups in merge_groups.items():
                # Remove processed maps from in_flight_maps
                for group in groups:
                    for partition in group:
                        del in_flight_maps[partition.id()]
                    total_size = sum(m.partition_metadata().size_bytes or 0 for m in group)
                    merge_step = PartitionTaskBuilder[PartitionT](
                        inputs=[p.partition() for p in group],
                        partial_metadatas=[m.partition_metadata() for m in group],
                        resource_request=ResourceRequest(memory_bytes=total_size),
                        node_id=node_id,
                    ).add_instruction(instruction=execution_step.ReduceMerge())
                    yield merge_step

        # Process next map task if available
        try:
            child_step = next(map_plan)
            if isinstance(child_step, PartitionTaskBuilder):
                child_step = child_step.finalize_partition_task_single_output(stage_id=stage_id)
                in_flight_maps[child_step.id()] = child_step
            yield child_step

        except StopIteration:
            no_more_input = True
            if in_flight_maps:
                logger.debug("PreShuffleMerge blocked on completion of a map task")
                yield None
            else:
                break



"""Given a commit hash and a "platform substring", prints the wheelname of the wheel (if one exists) to stdout.

# Example

```bash
COMMIT_HASH="abcdef0123456789"
PLATFORM_SUBSTRING="x86"
WHEELNAME=$(python get_wheel_name_from_s3.py $COMMIT_HASH $PLATFORM_SUBSTRING)

echo $WHEELNAME
# Will echo the wheelname if a wheel exists that matches the platform substring.
# Otherwise, will echo nothing.
```
"""

import argparse
from pathlib import Path

import boto3
import wheellib

if __name__ == "__main__":
    parser = argparse.ArgumentParser()
    parser.add_argument("--commit-hash", required=True)
    parser.add_argument("--platform-substring", required=True, choices=["x86", "aarch", "arm"])
    args = parser.parse_args()

    commit_hash = args.commit_hash
    platform_substring = args.platform_substring

    s3 = boto3.client("s3")
    response = s3.list_objects_v2(Bucket="github-actions-artifacts-bucket", Prefix=f"builds/{commit_hash}/")
    matches = []
    for content in response.get("Contents", []):
        wheelname = Path(content["Key"]).name
        platform_tag = wheellib.get_platform_tag(wheelname)
        if platform_substring in platform_tag:
            matches.append(wheelname)

    if len(matches) > 1:
        raise RuntimeError(
            f"Multiple wheels found that match the given platform substring: {platform_substring}; expected just 1"
        )

    print(matches[0]) if matches else None



from __future__ import annotations

from abc import abstractmethod
from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from daft.daft import FileFormatConfig, FileInfos, IOConfig


class RunnerIO:
    """Reading and writing data from the Runner.

    This is an abstract class and each runner must write their own implementation.
    """

    @abstractmethod
    def glob_paths_details(
        self,
        source_path: list[str],
        file_format_config: FileFormatConfig | None = None,
        io_config: IOConfig | None = None,
    ) -> FileInfos:
        """Globs the specified filepath to construct a FileInfos object containing file and dir metadata.

        Args:
            source_path (str): path to glob

        Returns:
            FileInfo: The file infos for the globbed paths.
        """
        raise NotImplementedError()



from typing import Dict, List, Optional, Union

from daft import Series
from daft.expressions import ExpressionsProjection
from daft.recordbatch.recordbatch import RecordBatch

from .micropartition import MicroPartition


def partition_strings_to_path(
    root_path: str,
    parts: Dict[str, str],
    partition_null_fallback: str = "__HIVE_DEFAULT_PARTITION__",
) -> str:
    keys = parts.keys()
    values = [partition_null_fallback if value is None else value for value in parts.values()]
    postfix = "/".join(f"{k}={v}" for k, v in zip(keys, values))
    return f"{root_path}/{postfix}"


def partition_values_to_str_mapping(
    partition_values: Union[MicroPartition, RecordBatch],
) -> Dict[str, Series]:
    null_part = Series.from_pylist(
        [None]
    )  # This is to ensure that the null values are replaced with the default_partition_fallback value
    pkey_names = partition_values.column_names()

    partition_strings = {}

    for c in pkey_names:
        column = partition_values.get_column(c)
        string_names = column._to_str_values()
        null_filled = column.is_null().if_else(null_part, string_names)
        partition_strings[c] = null_filled

    return partition_strings


class PartitionedTable:
    def __init__(self, table: MicroPartition, partition_keys: Optional[ExpressionsProjection]):
        self.table = table
        self.partition_keys = partition_keys
        self._partitions = None
        self._partition_values = None

    def _create_partitions(self):
        if self.partition_keys is None or len(self.partition_keys) == 0:
            self._partitions = [self.table]
            self._partition_values = None
        else:
            self._partitions, self._partition_values = self.table.partition_by_value(partition_keys=self.partition_keys)

    def partitions(self) -> List[MicroPartition]:
        """Returns a list of MicroPartitions representing the table partitioned by the partition keys.

        If the table is not partitioned, returns the original table as the single element in the list.
        """
        if self._partitions is None:
            self._create_partitions()
        return self._partitions  # type: ignore

    def partition_values(self) -> Optional[MicroPartition]:
        """Returns the partition values, with each row corresponding to the partition at the same index in PartitionedTable.partitions().

        If the table is not partitioned, returns None.

        """
        if self._partition_values is None:
            self._create_partitions()
        return self._partition_values

    def partition_values_str(self) -> Optional[MicroPartition]:
        """Returns the partition values converted to human-readable strings, keeping null values as null.

        If the table is not partitioned, returns None.
        """
        partition_values = self.partition_values()

        if partition_values is None:
            return None
        else:
            partition_strings = partition_values_to_str_mapping(partition_values)
            return MicroPartition.from_pydict(partition_strings)



import pyarrow as pa
import pytest

from daft.datatype import DataType
from daft.expressions import col
from daft.recordbatch import MicroPartition


def test_list_chunk_empty_series():
    table = MicroPartition.from_pydict(
        {
            "col": pa.array([], type=pa.list_(pa.int64())),
            "fixed_col": pa.array([], type=pa.list_(pa.int64(), 2)),
        }
    )

    result = table.eval_expression_list(
        [
            col("col").list.chunk(1).alias("col1"),
            col("col").list.chunk(2).alias("col2"),
            col("col").list.chunk(1000).alias("col3"),
            col("fixed_col").list.chunk(1).alias("fixed_col1"),
            col("fixed_col").list.chunk(2).alias("fixed_col2"),
            col("fixed_col").list.chunk(1000).alias("fixed_col3"),
        ]
    )

    assert result.to_pydict() == {
        "col1": [],
        "col2": [],
        "col3": [],
        "fixed_col1": [],
        "fixed_col2": [],
        "fixed_col3": [],
    }


def test_list_chunk():
    table = MicroPartition.from_pydict(
        {
            # Test list of an atomic type.
            "col1": [[1, 2, None, 4, 5], None, [7, 8, 9, None, 11, 12]],
            # Test lists of nested types.
            "col2": [
                [[1, 1], [2, 2], [None, None], [4, 4], [5, 5]],
                [[7, 7], [8, 8], [9, 9], [None, None], [11, 11], [12, 12]],
                None,
            ],
            "col3": [
                [[[1, 1]], None, [[None, None]], [[4, 4]], [[5, 5]]],
                [[[7, 7]], [[8, 8]], [None], [[None, None]], None, [[12, 12]]],
                None,
            ],
        }
    )

    result = table.eval_expression_list(
        [
            col("col1").list.chunk(1).alias("col1-1"),
            col("col1").list.chunk(2).alias("col1-2"),
            col("col1").list.chunk(3).alias("col1-3"),
            col("col1").list.chunk(10).alias("col1-10"),  # Test chunk size > list size
            col("col2").list.chunk(2).alias("col2-2"),
            col("col3").list.chunk(2).alias("col3-2"),
        ]
    )

    assert result.to_pydict() == {
        "col1-1": [[[1], [2], [None], [4], [5]], None, [[7], [8], [9], [None], [11], [12]]],
        "col1-2": [[[1, 2], [None, 4]], None, [[7, 8], [9, None], [11, 12]]],
        "col1-3": [[[1, 2, None]], None, [[7, 8, 9], [None, 11, 12]]],
        "col1-10": [[], None, []],
        "col2-2": [
            [[[1, 1], [2, 2]], [[None, None], [4, 4]]],
            [[[7, 7], [8, 8]], [[9, 9], [None, None]], [[11, 11], [12, 12]]],
            None,
        ],
        "col3-2": [
            [[[[1, 1]], None], [[[None, None]], [[4, 4]]]],
            [[[[7, 7]], [[8, 8]]], [[None], [[None, None]]], [None, [[12, 12]]]],
            None,
        ],
    }


def test_fixed_size_list_chunk():
    table = MicroPartition.from_pydict(
        {
            # Test list of atomic types.
            "col1": [["aa", "bb", None, "dd"], ["ee", None, "gg", "hh"], ["ii", "jj", "kk", "ll"]],
            # Test list with invalid elements.
            "col2": [["aa", "bb", "cc", "dd"], None, ["ee", "ff", "gg", "hh"]],
            # Test list of a nested type.
            "col3": [
                [[1], [2]],
                [[None, 111], [22, 222]],
                None,
            ],
            # Test deeper nested types.
            "col4": [
                [[[1]], [[None]]],
                [[[None], [111]], None],
                None,
            ],
        }
    )

    dtype1 = DataType.fixed_size_list(DataType.string(), 4)
    dtype2 = DataType.fixed_size_list(DataType.list(DataType.int32()), 2)
    dtype3 = DataType.fixed_size_list(DataType.list(DataType.list(DataType.int32())), 2)

    table = table.eval_expression_list(
        [
            col("col1").cast(dtype1),
            col("col2").cast(dtype1),
            col("col3").cast(dtype2),
            col("col4").cast(dtype3),
        ]
    )

    result = table.eval_expression_list(
        [
            col("col1").list.chunk(2).alias("col1-2"),
            col("col1").list.chunk(3).alias("col1-3"),
            col("col2").list.chunk(2).alias("col2-2"),
            col("col2").list.chunk(3).alias("col2-3"),
            col("col3").list.chunk(2).alias("col3-2"),
            col("col3").list.chunk(3).alias("col3-3"),  # Test chunk size > list size
            col("col3").list.chunk(4).alias("col3-4"),  # Test chunk size > list size
            col("col4").list.chunk(2).alias("col4-2"),
        ]
    )

    assert result.to_pydict() == {
        "col1-2": [[["aa", "bb"], [None, "dd"]], [["ee", None], ["gg", "hh"]], [["ii", "jj"], ["kk", "ll"]]],
        "col1-3": [[["aa", "bb", None]], [["ee", None, "gg"]], [["ii", "jj", "kk"]]],
        "col2-2": [[["aa", "bb"], ["cc", "dd"]], None, [["ee", "ff"], ["gg", "hh"]]],
        "col2-3": [[["aa", "bb", "cc"]], None, [["ee", "ff", "gg"]]],
        "col3-2": [[[[1], [2]]], [[[None, 111], [22, 222]]], None],
        "col3-3": [[], [], None],
        "col3-4": [[], [], None],
        "col4-2": [[[[[1]], [[None]]]], [[[[None], [111]], None]], None],
    }


def test_list_chunk_invalid_parameters():
    table = MicroPartition.from_pydict(
        {
            "col": [["a", "b", "c"], ["aa", "bb", "cc"], None, [None, "bbbb"], ["aaaaa", None]],
            "size": [1, 2, 3, 2, 2],
        }
    )
    with pytest.raises(ValueError, match="Invalid value for `size`"):
        table.eval_expression_list([col("col").list.chunk(col("size"))])
    with pytest.raises(ValueError, match="Invalid value for `size`"):
        table.eval_expression_list([col("col").list.chunk(0)])
    with pytest.raises(ValueError, match="Invalid value for `size`"):
        table.eval_expression_list([col("col").list.chunk(-1)])
    with pytest.raises(ValueError, match="Invalid value for `size`"):
        table.eval_expression_list([col("col").list.chunk(1.0)])


def test_list_chunk_non_list_type():
    table = MicroPartition.from_pydict(
        {
            "structcol": [{"a": 1}, {"b": 1}, {"c": 1}],
            "stringcol": ["a", "b", "c"],
            "intcol": [1, 2, 3],
        },
    )

    with pytest.raises(ValueError):
        table.eval_expression_list([col("structcol").list.chunk(2)])
    with pytest.raises(ValueError):
        table.eval_expression_list([col("stringcol").list.chunk(2)])
    with pytest.raises(ValueError):
        table.eval_expression_list([col("intcol").list.chunk(2)])



from __future__ import annotations

import pytest

import daft
from daft.expressions import col
from tests.conftest import assert_df_equals

COL_SUBSET = ["Unique Key", "Complaint Type", "Borough", "Descriptor"]


@pytest.mark.parametrize(
    "daft_df_ops",
    [
        pytest.param(
            lambda daft_df: (
                daft_df.where(col("Complaint Type") == "Noise - Street/Sidewalk").select(
                    col("Unique Key"),
                    col("Complaint Type"),
                    col("Borough"),
                    col("Descriptor"),
                )
            ),
            id="where..select",
        ),
        pytest.param(
            lambda daft_df: (
                daft_df.select(
                    col("Unique Key"),
                    col("Complaint Type"),
                    col("Borough"),
                    col("Descriptor"),
                ).where(col("Complaint Type") == "Noise - Street/Sidewalk")
            ),
            id="select..where",
        ),
    ],
)
def test_filter(daft_df_ops, daft_df, service_requests_csv_pd_df, repartition_nparts, with_morsel_size):
    """Filter the dataframe, retrieve the top N results and select a subset of columns."""
    daft_noise_complaints = daft_df_ops(daft_df.repartition(repartition_nparts))

    pd_noise_complaints = service_requests_csv_pd_df[
        service_requests_csv_pd_df["Complaint Type"] == "Noise - Street/Sidewalk"
    ][COL_SUBSET]
    daft_pd_df = daft_noise_complaints.to_pandas()
    assert_df_equals(daft_pd_df, pd_noise_complaints)


@pytest.mark.parametrize(
    "daft_df_ops",
    [
        pytest.param(
            lambda daft_df: (
                daft_df.where(
                    (
                        (col("Complaint Type") == "Noise - Street/Sidewalk")
                        | (col("Complaint Type") == "Noise - Commercial")
                    )
                    & (col("Borough") == "BROOKLYN")
                ).select(
                    col("Unique Key"),
                    col("Complaint Type"),
                    col("Borough"),
                    col("Descriptor"),
                )
            ),
            id="where..select",
        ),
        pytest.param(
            lambda daft_df: (
                daft_df.select(
                    col("Unique Key"),
                    col("Complaint Type"),
                    col("Borough"),
                    col("Descriptor"),
                ).where(
                    (
                        (col("Complaint Type") == "Noise - Street/Sidewalk")
                        | (col("Complaint Type") == "Noise - Commercial")
                    )
                    & (col("Borough") == "BROOKLYN")
                )
            ),
            id="select..where",
        ),
        pytest.param(
            lambda daft_df: (
                daft_df.select(
                    col("Unique Key"),
                    col("Complaint Type"),
                    col("Borough"),
                    col("Descriptor"),
                ).where(
                    (col("Borough") == "BROOKLYN")
                    & (
                        (col("Complaint Type") == "Noise - Street/Sidewalk")
                        | (col("Complaint Type") == "Noise - Commercial")
                    )
                )
            ),
            id="select..where(flipped&)",
        ),
    ],
)
def test_complex_filter(daft_df_ops, daft_df, service_requests_csv_pd_df, repartition_nparts, with_morsel_size):
    """Filter the dataframe with a complex filter and select a subset of columns."""
    daft_noise_complaints_brooklyn = daft_df_ops(daft_df.repartition(repartition_nparts))

    pd_noise_complaints_brooklyn = service_requests_csv_pd_df[
        (
            (service_requests_csv_pd_df["Complaint Type"] == "Noise - Street/Sidewalk")
            | (service_requests_csv_pd_df["Complaint Type"] == "Noise - Commercial")
        )
        & (service_requests_csv_pd_df["Borough"] == "BROOKLYN")
    ][COL_SUBSET]
    daft_pd_df = daft_noise_complaints_brooklyn.to_pandas()
    assert_df_equals(daft_pd_df, pd_noise_complaints_brooklyn)


@pytest.mark.parametrize(
    "daft_df_ops",
    [
        pytest.param(
            lambda daft_df: (
                daft_df.where(col("Complaint Type") == "Noise - Street/Sidewalk")
                .where(col("Borough") == "BROOKLYN")
                .select(
                    col("Unique Key"),
                    col("Complaint Type"),
                    col("Borough"),
                    col("Descriptor"),
                )
            ),
            id="where..where..select",
        ),
        pytest.param(
            lambda daft_df: (
                daft_df.where(col("Complaint Type") == "Noise - Street/Sidewalk")
                .select(
                    col("Unique Key"),
                    col("Complaint Type"),
                    col("Borough"),
                    col("Descriptor"),
                )
                .where(col("Borough") == "BROOKLYN")
            ),
            id="where..select..where",
        ),
        pytest.param(
            lambda daft_df: (
                daft_df.select(
                    col("Unique Key"),
                    col("Complaint Type"),
                    col("Borough"),
                    col("Descriptor"),
                )
                .where(col("Complaint Type") == "Noise - Street/Sidewalk")
                .where(col("Borough") == "BROOKLYN")
            ),
            id="select..where..where",
        ),
    ],
)
def test_chain_filter(daft_df_ops, daft_df, service_requests_csv_pd_df, repartition_nparts, with_morsel_size):
    """Filter the dataframe with a chain of filters and select a subset of columns."""
    daft_noise_complaints_brooklyn = daft_df_ops(daft_df.repartition(repartition_nparts))

    pd_noise_complaints_brooklyn = service_requests_csv_pd_df
    pd_noise_complaints_brooklyn = pd_noise_complaints_brooklyn[
        pd_noise_complaints_brooklyn["Complaint Type"] == "Noise - Street/Sidewalk"
    ]
    pd_noise_complaints_brooklyn = pd_noise_complaints_brooklyn[pd_noise_complaints_brooklyn["Borough"] == "BROOKLYN"][
        COL_SUBSET
    ]
    daft_pd_df = daft_noise_complaints_brooklyn.to_pandas()
    assert_df_equals(daft_pd_df, pd_noise_complaints_brooklyn)


def test_filter_on_projection():
    """Filter the dataframe with on top of a projection."""
    df = daft.from_pydict({"x": [1, 1, 1, 1, 1]})
    df = df.select(col("x") * 2)
    df = df.where(col("x") == 1)
    result = df.to_pandas()
    assert len(result) == 0

    df = daft.from_pydict({"x": [1, 1, 1, 1, 1]})
    df = df.select(col("x") * 2)
    df = df.where(col("x") == 2)
    result = df.to_pandas()
    assert len(result) == 5
    assert (result == 2).all().all()



from __future__ import annotations

from itertools import product

import pytest

import daft
from daft.daft import CsvConvertOptions
from daft.expressions.expressions import _resolved_col
from daft.recordbatch import MicroPartition

PRED_PUSHDOWN_FILES = [
    "s3://daft-public-data/test_fixtures/csv-dev/sampled-tpch.csv",
    "tests/assets/sampled-tpch.csv",
]


# @pytest.mark.integration()
@pytest.mark.parametrize(
    "path, pred, limit",
    product(
        PRED_PUSHDOWN_FILES,
        [_resolved_col("L_ORDERKEY") == 7, _resolved_col("L_ORDERKEY") == 10000, daft.lit(True)],
        [None, 1, 1000],
    ),
)
def test_csv_filter_pushdowns(path, pred, limit, aws_public_s3_config):
    with_pushdown = MicroPartition.read_csv(
        path,
        io_config=aws_public_s3_config,
        convert_options=CsvConvertOptions(limit=limit, predicate=pred._expr),
        parse_options=None,
        read_options=None,
    )
    after = MicroPartition.read_csv(
        path, io_config=aws_public_s3_config, convert_options=None, parse_options=None, read_options=None
    ).filter([pred])
    if limit is not None:
        after = after.head(limit)
    assert with_pushdown.to_arrow() == after.to_arrow()


# @pytest.mark.integration()
@pytest.mark.parametrize(
    "path, pred",
    product(
        PRED_PUSHDOWN_FILES, [_resolved_col("L_ORDERKEY") == 7, _resolved_col("L_ORDERKEY") == 10000, daft.lit(True)]
    ),
)
def test_csv_filter_pushdowns_disjoint_predicate(path, pred, aws_public_s3_config):
    with_pushdown = MicroPartition.read_csv(
        path,
        convert_options=CsvConvertOptions(predicate=pred._expr, include_columns=["L_QUANTITY"]),
        parse_options=None,
        read_options=None,
        io_config=aws_public_s3_config,
    )
    after = (
        MicroPartition.read_csv(
            path, io_config=aws_public_s3_config, convert_options=None, parse_options=None, read_options=None
        )
        .filter([pred])
        .eval_expression_list([_resolved_col("L_QUANTITY")])
    )
    assert with_pushdown.to_arrow() == after.to_arrow()



from __future__ import annotations


def test_set_operation(spark_session):
    """Test the Set operation with various data types and edge cases."""
    configs = {
        "spark.test.string": "test_value",
        # todo: I check if non-strings are supported.
        # "spark.test.boolean": True,
        # "spark.test.integer": 42,
        # "spark.test.float": 3.14,
        "spark.test.empty": "",  # Test empty string
        "spark.test.special": "!@#$%^&*()",  # Test special characters
    }

    # Set all configurations
    for key, value in configs.items():
        spark_session.conf.set(key, value)

    # Verify all configurations
    for key, value in configs.items():
        assert str(spark_session.conf.get(key)) == str(value)


def test_get_operations(spark_session):
    """Test various Get operations including Get, GetWithDefault, and GetOption."""
    # Setup test data
    test_data = {"spark.test.existing": "value", "spark.prefix.one": "1", "spark.prefix.two": "2"}
    for key, value in test_data.items():
        spark_session.conf.set(key, value)

    # Test basic Get
    assert spark_session.conf.get("spark.test.existing") == "value"

    # Test GetWithDefault
    assert spark_session.conf.get("spark.test.nonexistent", "default") == "default"
    assert spark_session.conf.get("spark.test.existing", "default") == "value"

    # Test GetOption (if implemented)
    # Note: This might need to be adjusted based on actual GetOption implementation
    assert spark_session.conf.get("spark.test.nonexistent") is None

    # Test GetAll with prefix
    prefix_configs = {key: spark_session.conf.get(key) for key in ["spark.prefix.one", "spark.prefix.two"]}
    assert prefix_configs == {"spark.prefix.one": "1", "spark.prefix.two": "2"}


def test_unset_operation(spark_session):
    """Test the Unset operation with various scenarios."""
    # Setup test data
    spark_session.conf.set("spark.test.temp", "value")

    # Test basic unset
    spark_session.conf.unset("spark.test.temp")
    assert spark_session.conf.get("spark.test.temp") is None

    # Test unset non-existent key (should not raise error)
    spark_session.conf.unset("spark.test.nonexistent")

    # Test unset and then set again
    key = "spark.test.resettable"
    spark_session.conf.set(key, "first")
    spark_session.conf.unset(key)
    spark_session.conf.set(key, "second")
    assert spark_session.conf.get(key) == "second"


def test_edge_cases(spark_session):
    """Test various edge cases and potential error conditions."""
    # Test very long key and value
    long_key = "spark.test." + "x" * 1000
    long_value = "y" * 1000
    spark_session.conf.set(long_key, long_value)
    assert spark_session.conf.get(long_key) == long_value

    # Test unicode characters
    unicode_key = "spark.test.unicode"
    unicode_value = "测试值"
    spark_session.conf.set(unicode_key, unicode_value)
    assert spark_session.conf.get(unicode_key) == unicode_value

    # Test setting same key multiple times rapidly
    key = "spark.test.rapid"
    for i in range(100):
        spark_session.conf.set(key, f"value{i}")
    assert spark_session.conf.get(key) == "value99"

    # Test concurrent modifications (if supported)
    # Note: This might need to be adjusted based on concurrency support
    from concurrent.futures import ThreadPoolExecutor

    key = "spark.test.concurrent"

    def modify_conf(i: int):
        spark_session.conf.set(key, f"value{i}")

    with ThreadPoolExecutor(max_workers=4) as executor:
        list(executor.map(modify_conf, range(100)))

    assert spark_session.conf.get(key) is not None  # Value should be set to something



# /// script
# dependencies = ['numpy']
# ///

import argparse
import random
import time
from functools import partial
from typing import Any, Dict, Optional

import numpy as np
import pyarrow as pa

import daft
import daft.context
from daft.io._generator import read_generator
from daft.recordbatch.recordbatch import RecordBatch

# Constants
GB = 1 << 30
MB = 1 << 20
KB = 1 << 10
ROW_SIZE = 100 * KB


def parse_size(size_str: str) -> int:
    """Convert human-readable size string to bytes."""
    units = {"B": 1, "KB": KB, "MB": MB, "GB": GB}
    size_str = size_str.upper()
    value, unit = size_str.split(" ")
    return int(float(value) * units[unit])


def get_skewed_distribution(num_partitions: int, skew_factor: float) -> np.ndarray:
    """Generate a skewed distribution using a power law.

    Higher skew_factor means more skewed distribution.
    """
    if skew_factor <= 0:
        return np.ones(num_partitions) / num_partitions

    # Generate power law distribution
    x = np.arange(1, num_partitions + 1)
    weights = 1.0 / (x**skew_factor)
    return weights / weights.sum()


def get_partition_size(base_size: int, size_variation: float, partition_idx: int) -> int:
    """Calculate size for a specific partition with variation.

    Args:
        base_size: The base partition size in bytes
        size_variation: Float between 0 and 1 indicating maximum variation (e.g., 0.2 = ±20%)
        partition_idx: Index of the partition (used for random seed)

    Returns:
        Adjusted partition size in bytes
    """
    if size_variation <= 0:
        return base_size

    # Use partition_idx as seed for consistent variation per partition
    random.seed(f"size_{partition_idx}")

    # Generate variation factor between (1-variation) and (1+variation)
    variation_factor = 1.0 + random.uniform(-size_variation, size_variation)

    # Ensure we don't go below 10% of base size
    min_size = base_size * 0.1
    return max(int(base_size * variation_factor), int(min_size))


def generate(
    num_partitions: int,
    base_partition_size: int,
    skew_factor: float,
    timing_variation: float,
    size_variation: float,
    partition_idx: int,
):
    """Generate data for a single partition with optional skew, timing and size variations."""
    # Calculate actual partition size with variation
    actual_partition_size = get_partition_size(base_partition_size, size_variation, partition_idx)
    num_rows = actual_partition_size // ROW_SIZE

    # Apply skewed distribution if specified
    if skew_factor > 0:
        weights = get_skewed_distribution(num_partitions, skew_factor)
        data = {
            "ints": pa.array(
                np.random.choice(num_partitions, size=num_rows, p=weights, replace=True).astype(np.uint64),
            ),
            "bytes": pa.array(
                [np.random.bytes(ROW_SIZE) for _ in range(num_rows)],
                type=pa.binary(ROW_SIZE),
            ),
        }
    else:
        data = {
            "ints": pa.array(np.random.randint(0, num_partitions, size=num_rows)),
            "bytes": pa.array(
                [np.random.bytes(ROW_SIZE) for _ in range(num_rows)],
                type=pa.binary(ROW_SIZE),
            ),
        }

    # Simulate varying processing times if specified
    if timing_variation > 0:
        random.seed(f"timing_{partition_idx}")
        delay = random.uniform(0, timing_variation)
        time.sleep(delay)

    yield RecordBatch.from_pydict(data)


def generator(
    num_partitions: int,
    partition_size: int,
    skew_factor: float,
    timing_variation: float,
    size_variation: float,
):
    """Generate data for all partitions."""
    for i in range(num_partitions):
        yield partial(
            generate,
            num_partitions,
            partition_size,
            skew_factor,
            timing_variation,
            size_variation,
            i,
        )


def setup_daft(shuffle_algorithm: Optional[str] = None):
    """Configure Daft execution settings."""
    daft.context.set_runner_ray()
    daft.context.set_execution_config(shuffle_algorithm=shuffle_algorithm, pre_shuffle_merge_threshold=8 * GB)


def create_schema():
    """Create the Daft schema for the dataset."""
    return daft.Schema._from_field_name_and_types([("ints", daft.DataType.uint64()), ("bytes", daft.DataType.binary())])


def run_benchmark(
    num_partitions: int,
    partition_size: int,
    skew_factor: float,
    timing_variation: float,
    size_variation: float,
    shuffle_algorithm: Optional[str] = None,
) -> Dict[str, Any]:
    """Run the memory benchmark and return statistics."""
    setup_daft(shuffle_algorithm)
    schema = create_schema()

    def benchmark_func():
        return (
            read_generator(
                generator(
                    num_partitions,
                    partition_size,
                    skew_factor,
                    timing_variation,
                    size_variation,
                ),
                schema,
            )
            .repartition(num_partitions, "ints")
            .collect()
        )

    start_time = time.time()

    benchmark_func()

    end_time = time.time()
    return end_time - start_time


def main():
    parser = argparse.ArgumentParser(description="Run memory benchmark for data processing")
    parser.add_argument("--partitions", type=int, default=1000, help="Number of partitions")
    parser.add_argument(
        "--partition-size",
        type=str,
        default="100 MB",
        help="Base size for each partition (e.g., 300 MB, 1 GB)",
    )
    parser.add_argument(
        "--skew-factor",
        type=float,
        default=0.0,
        help="Skew factor for partition distribution (0.0 for uniform, higher for more skew)",
    )
    parser.add_argument(
        "--timing-variation",
        type=float,
        default=0.0,
        help="Maximum random delay in seconds for partition processing",
    )
    parser.add_argument(
        "--size-variation",
        type=float,
        default=0.0,
        help="Maximum partition size variation as fraction (0.0-1.0, e.g., 0.2 for ±20%%)",
    )
    parser.add_argument("--shuffle-algorithm", type=str, default=None, help="Shuffle algorithm to use")

    args = parser.parse_args()

    if not 0 <= args.size_variation <= 1:
        parser.error("Size variation must be between 0 and 1")

    partition_size_bytes = parse_size(args.partition_size)

    print("Running benchmark with configuration:")
    print(f"Partitions: {args.partitions}")
    print(f"Base partition size: {args.partition_size} ({partition_size_bytes} bytes)")
    print(f"Size variation: ±{args.size_variation*100:.0f}%")
    print(f"Row size: {ROW_SIZE/KB:.0f}KB (fixed)")
    print(f"Skew factor: {args.skew_factor}")
    print(f"Timing variation: {args.timing_variation}s")
    print(f"Shuffle algorithm: {args.shuffle_algorithm or 'default'}")

    try:
        timing = run_benchmark(
            num_partitions=args.partitions,
            partition_size=partition_size_bytes,
            skew_factor=args.skew_factor,
            timing_variation=args.timing_variation,
            size_variation=args.size_variation,
            shuffle_algorithm=args.shuffle_algorithm,
        )

        print("\nRan benchmark with configuration:")
        print(f"Partitions: {args.partitions}")
        print(f"Base partition size: {args.partition_size} ({partition_size_bytes} bytes)")
        print(f"Size variation: ±{args.size_variation*100:.0f}%")
        print(f"Row size: {ROW_SIZE/KB:.0f}KB (fixed)")
        print(f"Skew factor: {args.skew_factor}")
        print(f"Timing variation: {args.timing_variation}s")
        print(f"Shuffle algorithm: {args.shuffle_algorithm or 'default'}")
        print("\nResults:")
        print(f"Total time: {timing:.2f}s")

    except Exception as e:
        print(f"Error running benchmark: {e!s}")
        raise


if __name__ == "__main__":
    main()



import daft
import daft.recordbatch
from daft import DataFrame, Schema
from daft import DataType as dt
from daft.logical.schema import Field


def test_describe():
    fields = [
        Field.create("c_int8", dt.int8()),
        Field.create("c_int16", dt.int16()),
        Field.create("c_int32", dt.int32()),
        Field.create("c_int64", dt.int64()),
        Field.create("c_uint8", dt.uint8()),
        Field.create("c_uint16", dt.uint16()),
        Field.create("c_uint32", dt.uint32()),
        Field.create("c_uint64", dt.uint64()),
        Field.create("c_float32", dt.float32()),
        Field.create("c_float64", dt.float64()),
        Field.create("c_string", dt.string()),
        Field.create("c_bool", dt.bool()),
        Field.create("c_binary", dt.binary()),
        Field.create("c_date", dt.date()),
        Field.create("c_time", dt.time(daft.TimeUnit.ns())),
        Field.create("c_timestamp", dt.timestamp(daft.TimeUnit.ns())),
        Field.create("c_duration", dt.duration(daft.TimeUnit.ns())),
        Field.create("c_interval", dt.interval()),
        Field.create("c_list", dt.list(dt.int32())),
        Field.create("c_fixed_size_list", dt.fixed_size_list(dt.int32(), 2)),
        Field.create("c_map", dt.map(dt.string(), dt.int32())),
        Field.create("c_struct", dt.struct({"a": dt.int32()})),
        Field.create("c_extension", dt.extension("test", dt.int32())),
        Field.create("c_embedding", dt.embedding(dt.float32(), 4)),
        Field.create("c_image", dt.image()),
        Field.create("c_tensor", dt.tensor(dt.float32())),
        Field.create("c_sparse_tensor", dt.sparse_tensor(dt.float32())),
    ]
    # create an empty table with known schema
    mp = daft.recordbatch.MicroPartition.empty(Schema._from_fields(fields))
    df = DataFrame._from_tables(mp)
    # describe..
    df = df.describe().collect()
    # assert
    assert df.count_rows() == len(fields), "Output from describe() is missing some columns"
    expect_columns = set(f.name for f in fields)
    actual_columns = df.to_pydict()["column_name"]
    for c in actual_columns:
        assert c in expect_columns, f"Column {c} from describe() not found in original schema"



from __future__ import annotations

import os
import socket
import urllib
from unittest.mock import MagicMock, patch

from daft.scarf_telemetry import scarf_telemetry


@patch("daft.scarf_telemetry.get_build_type")
@patch("daft.scarf_telemetry.get_version")
@patch("urllib.request.urlopen")
def test_scarf_telemetry_basic(mock_urlopen: MagicMock, mock_version: MagicMock, mock_build_type: MagicMock):
    # Test basic functionality of scarf_telemetry verify that analytics are successfully sent and url is properly formatted with all required parameters

    # Set up mocks for version and build_type
    mock_version.return_value = "0.0.0"
    mock_build_type.return_value = "release"
    mock_response = MagicMock()
    mock_response.status = 200
    mock_urlopen.return_value.__enter__.return_value = mock_response

    # Test basic analytics call
    response_status, runner_type = scarf_telemetry(runner="ray")

    assert response_status == "Response status: 200"
    assert runner_type == "ray"

    # Verify URL format and parameters
    called_url = mock_urlopen.call_args[0][0]
    assert called_url.startswith("https://daft.gateway.scarf.sh/daft-runner?")
    assert "version=0.0.0" in called_url
    assert "runner=ray" in called_url


@patch("daft.scarf_telemetry.get_build_type")
@patch("daft.scarf_telemetry.get_version")
def test_scarf_telemetry_dev_build(mock_version: MagicMock, mock_build_type: MagicMock):
    # Test that analytics are not sent for dev builds, function returns None for both status and runner type

    mock_version.return_value = "0.0.0"
    mock_build_type.return_value = "dev"

    response_status, runner_type = scarf_telemetry(runner="ray")

    assert response_status is None
    assert runner_type is None


@patch("daft.scarf_telemetry.get_build_type")
@patch("daft.scarf_telemetry.get_version")
@patch("urllib.request.urlopen")
def test_scarf_telemetry_opt_out_with_scarf_analytics(
    mock_urlopen: MagicMock,
    mock_version: MagicMock,
    mock_build_type: MagicMock,
):
    mock_version.return_value = "0.0.0"
    mock_build_type.return_value = "release"

    # Clean and set environment
    os.environ.pop("DO_NOT_TRACK", None)
    os.environ["SCARF_NO_ANALYTICS"] = "true"

    response_status, runner_type = scarf_telemetry(runner="ray")

    assert response_status is None
    assert runner_type is None
    mock_urlopen.assert_not_called()


@patch("daft.scarf_telemetry.get_build_type")
@patch("daft.scarf_telemetry.get_version")
@patch("urllib.request.urlopen")
def test_scarf_telemetry_opt_out_with_do_not_track(
    mock_urlopen: MagicMock,
    mock_version: MagicMock,
    mock_build_type: MagicMock,
):
    mock_version.return_value = "0.0.0"
    mock_build_type.return_value = "release"

    # Clean and set environment
    os.environ.pop("SCARF_NO_ANALYTICS", None)
    os.environ["DO_NOT_TRACK"] = "true"

    response_status, runner_type = scarf_telemetry(runner="ray")

    assert response_status is None
    assert runner_type is None
    mock_urlopen.assert_not_called()


@patch("daft.scarf_telemetry.get_build_type")
@patch("daft.scarf_telemetry.get_version")
@patch("urllib.request.urlopen")
def test_scarf_telemetry_error_handling(
    mock_urlopen: MagicMock,
    mock_version: MagicMock,
    mock_build_type: MagicMock,
):
    # Test error handling in scarf_telemetry, verifies that network errors are caught, function returns error message and None for runner type

    # Clean environment
    os.environ.pop("SCARF_NO_ANALYTICS", None)
    os.environ.pop("DO_NOT_TRACK", None)

    mock_version.return_value = "0.0.0"
    mock_build_type.return_value = "release"
    mock_urlopen.side_effect = urllib.error.URLError(socket.timeout("Timeout"))

    response_status, runner_type = scarf_telemetry(runner="ray")

    assert response_status.startswith("Analytics error:")
    assert runner_type is None



import daft


def test_to_pylist() -> None:
    df = daft.from_pydict({"a": [1, 2, 3, 4], "b": [2, 4, 3, 1]})
    res = df.to_pylist()
    assert res == [{"a": 1, "b": 2}, {"a": 2, "b": 4}, {"a": 3, "b": 3}, {"a": 4, "b": 1}]


def test_to_pylist_with_None() -> None:
    df = daft.from_pydict({"a": [None], "b": [None]})
    assert df.to_pylist() == [{"a": None, "b": None}]


def test_to_pylist_with_string() -> None:
    df = daft.from_pydict({"a": ["Hello"], "b": ["World"]})
    assert df.to_pylist() == [{"a": "Hello", "b": "World"}]


def test_to_pylist_with_int() -> None:
    df = daft.from_pydict({"a": [1, 2], "b": [3, 4]})
    assert df.to_pylist() == [{"a": 1, "b": 3}, {"a": 2, "b": 4}]


def test_to_pylist_with_int_None() -> None:
    df = daft.from_pydict({"a": [1, None, 2], "b": [3, 4, None]})
    assert df.to_pylist() == [{"a": 1, "b": 3}, {"a": None, "b": 4}, {"a": 2, "b": None}]


def test_to_pylist_with_float() -> None:
    df = daft.from_pydict({"a": [1.1, 2.2], "b": [3.3, 4.4]})
    assert df.to_pylist() == [{"a": 1.1, "b": 3.3}, {"a": 2.2, "b": 4.4}]


def test_to_pylist_with_float_None() -> None:
    df = daft.from_pydict({"a": [1.1, None, 2.2], "b": [3.3, 4.4, None]})
    assert df.to_pylist() == [{"a": 1.1, "b": 3.3}, {"a": None, "b": 4.4}, {"a": 2.2, "b": None}]


def test_to_pylist_with_empty() -> None:
    df = daft.from_pydict({"a": [], "b": []})
    assert df.to_pylist() == []



from __future__ import annotations

import os

###
# Set up code coverage for when running code coverage with ray
###
if "COV_CORE_SOURCE" in os.environ:
    try:
        from pytest_cov.embed import init

        init()
    except Exception as exc:
        import sys

        sys.stderr.write(
            "pytest-cov: Failed to setup subprocess coverage. " "Environ: {!r} " "Exception: {!r}\n".format(
                {k: v for k, v in os.environ.items() if k.startswith("COV_CORE")}, exc
            )
        )

###
# Get build constants from Rust .so
###

from daft.daft import build_type as _build_type
from daft.daft import version as _version
from daft.daft import refresh_logger as _refresh_logger


def get_version() -> str:
    return _version()


def get_build_type() -> str:
    return _build_type()


def refresh_logger() -> None:
    """Refreshes Daft's internal rust logging to the current python log level."""
    _refresh_logger()


__version__ = get_version()

###
# Initialize analytics
###

from daft.analytics import init_analytics

user_opted_out = os.getenv("DAFT_ANALYTICS_ENABLED") == "0"
analytics_client = init_analytics(get_version(), get_build_type(), user_opted_out)
analytics_client.track_import()

###
# Daft top-level imports
###

from daft.catalog import (
    Catalog,
    Identifier,
    Table,
    register_table,
)
from daft.context import set_execution_config, set_planning_config, execution_config_ctx, planning_config_ctx
from daft.convert import (
    from_arrow,
    from_dask_dataframe,
    from_pandas,
    from_pydict,
    from_pylist,
    from_ray_dataset,
)
from daft.daft import ImageFormat, ImageMode, ResourceRequest
from daft.dataframe import DataFrame
from daft.logical.schema import Schema
from daft.datatype import DataType, TimeUnit
from daft.expressions import Expression, col, list_, lit, interval, struct, coalesce
from daft.io import (
    DataCatalogTable,
    DataCatalogType,
    from_glob_path,
    read_csv,
    read_deltalake,
    read_hudi,
    read_iceberg,
    read_json,
    read_parquet,
    read_sql,
    read_lance,
)
from daft.series import Series
from daft.session import (
    Session,
    attach_catalog,
    attach_table,
    detach_catalog,
    detach_table,
    create_temp_table,
    current_catalog,
    current_namespace,
    current_session,
    get_catalog,
    get_table,
    has_catalog,
    has_table,
    list_catalogs,
    list_tables,
    read_table,
    set_catalog,
    set_namespace,
    set_session,
)
from daft.sql import sql, sql_expr
from daft.udf import udf
from daft.viz import register_viz_hook

to_struct = Expression.to_struct

__all__ = [
    "Catalog",
    "DataCatalogTable",
    "DataCatalogType",
    "DataFrame",
    "DataType",
    "Expression",
    "Identifier",
    "ImageFormat",
    "ImageMode",
    "ResourceRequest",
    "Schema",
    "Series",
    "Session",
    "Table",
    "TimeUnit",
    "attach_catalog",
    "attach_table",
    "coalesce",
    "col",
    "create_temp_table",
    "current_catalog",
    "current_namespace",
    "current_session",
    "detach_catalog",
    "detach_table",
    "execution_config_ctx",
    "from_arrow",
    "from_dask_dataframe",
    "from_glob_path",
    "from_pandas",
    "from_pydict",
    "from_pylist",
    "from_ray_dataset",
    "get_catalog",
    "get_table",
    "has_catalog",
    "has_table",
    "interval",
    "list_",
    "list_catalogs",
    "list_tables",
    "lit",
    "planning_config_ctx",
    "read_csv",
    "read_deltalake",
    "read_hudi",
    "read_iceberg",
    "read_json",
    "read_lance",
    "read_parquet",
    "read_sql",
    "read_table",
    "refresh_logger",
    "register_table",
    "register_viz_hook",
    "set_catalog",
    "set_execution_config",
    "set_namespace",
    "set_planning_config",
    "set_session",
    "sql",
    "sql_expr",
    "struct",
    "to_struct",
    "udf",
]



import argparse
import json
from datetime import datetime
from pathlib import Path

import daft
from daft.sql.sql import SQLCatalog

TABLE_NAMES = [
    "call_center",
    "catalog_page",
    "catalog_returns",
    "catalog_sales",
    "customer",
    "customer_address",
    "customer_demographics",
    "date_dim",
    "household_demographics",
    "income_band",
    "inventory",
    "item",
    "promotion",
    "reason",
    "ship_mode",
    "store",
    "store_returns",
    "store_sales",
    "time_dim",
    "warehouse",
    "web_page",
    "web_returns",
    "web_sales",
    "web_site",
]


def register_catalog(scale_factor: int) -> SQLCatalog:
    return SQLCatalog(
        tables={
            table: daft.read_parquet(
                f"s3://eventual-dev-benchmarking-fixtures/uncompressed/tpcds-dbgen/{scale_factor}/{table}"
            )
            for table in TABLE_NAMES
        }
    )


def run(
    question: int,
    dry_run: bool,
    scale_factor: int,
):
    catalog = register_catalog(scale_factor)
    query_file = Path(__file__).parent / "queries" / f"{question:02}.sql"
    with open(query_file) as f:
        query_string = f.read()

    info_path = Path("/tmp") / "ray" / "session_latest" / "logs" / "info"
    info_path.mkdir(parents=True, exist_ok=True)
    query = daft.sql(query_string, catalog=catalog)

    explain_delta = None
    with open(info_path / f"plan-{question}.txt", "w") as f:
        explain_start = datetime.now()
        query.explain(show_all=True, file=f, format="mermaid")
        explain_end = datetime.now()
        explain_delta = explain_end - explain_start

    execute_delta = None
    if not dry_run:
        execute_start = datetime.now()
        query.collect()
        execute_end = datetime.now()
        execute_delta = execute_end - execute_start

    with open(info_path / f"stats-{question}.txt", "w") as f:
        stats = json.dumps(
            {
                "question": question,
                "scale-factor": scale_factor,
                "planning-time": str(explain_delta),
                "execution-time": str(execute_delta),
            }
        )
        f.write(stats)


if __name__ == "__main__":
    parser = argparse.ArgumentParser()
    parser.add_argument(
        "--question",
        type=int,
        help="The TPC-DS question index to run",
        required=True,
    )
    parser.add_argument(
        "--dry-run",
        action="store_true",
        help="Whether or not to run the query in dry-run mode; if true, only the plan will be printed out",
    )
    parser.add_argument(
        "--scale-factor",
        type=int,
        help="Which scale factor to run this data at",
        required=False,
        default=2,
    )
    args = parser.parse_args()

    assert args.question in range(1, 100)

    run(question=args.question, dry_run=args.dry_run, scale_factor=args.scale_factor)
